### This Template code is needed to make the firmware compatible devicetypes and moduletypes likefied.
### It will aslo show a max of 3. Once you have more than 3, if will show you the amount and use the side page to show you all.

# FIRMWARE_DEVICE_TYPES = """
# {% load i18n %}
# {% if value.count > 3 %}
#   <a href="{% url 'dcim:devicetype_list' %}?firmware_id={{ record.pk }}">{{ value.count }} Device Types</a>
# {% else %}
#   {% for devicetype in value.all %}
#     <a href="{{ devicetype.get_absolute_url }}">{{ devicetype.model }}</a><br />
#   {% endfor %}
# {% endif %}
# """




### This Template code is needed to show all the firmware compatible devicetypes and moduletypes + make them likefied.


SHOW_LINKED_DEVICE_TYPES = """
{% load i18n %}
{% for devicetype in value.all %}
  <a href="{{ devicetype.get_absolute_url }}">{{ devicetype }}</a><br />
{% endfor %}
"""


SHOW_LINKED_MODULE_TYPES = """
{% load i18n %}
{% for moduletype in value.all %}
  <a href="{{ moduletype.get_absolute_url }}">{{ moduletype }}</a><br />
{% endfor %}
"""



import logging
from django.contrib import messages
from django.db import IntegrityError
from django.shortcuts import redirect
from django.template import Template
from netbox.views import generic
from utilities.views import register_model_view

from .. import filtersets
from .. import tables
from .. import forms
from .. import models

__all__ = (
    'FirmwareAssignmentView',
    'FirmwareAssignmentListView'
)

@register_model_view(models.FirmwareAssignment)
class FirmwareAssignmentView(generic.ObjectView):
    queryset = models.FirmwareAssignment.objects.all()

    def get_extra_context(self, request, instance):
        context = super().get_extra_context(request, instance)
        return context

@register_model_view(models.FirmwareAssignment, 'list', path='', detail=False)
class FirmwareAssignmentListView(generic.ObjectListView):
    queryset = models.FirmwareAssignment.objects.prefetch_related(
        'device',
        'module',
        'firmware',
    )
    filterset = filtersets.FirmwareAssignmentFilterSet
    filterset_form = forms.FirmwareAssignmentFilterForm
    table = tables.FirmwareAssignmentTable
    
@register_model_view(models.FirmwareAssignment, 'edit')
@register_model_view(models.FirmwareAssignment, 'add', detail=False)
class FirmwareAssignmentEditView(generic.ObjectEditView):
    queryset = models.FirmwareAssignment.objects.all()
    form = forms.FirmwareAssignmentForm
    default_return_url = 'plugins:netbox_firmware:firmwareassignment_list'

@register_model_view(models.FirmwareAssignment,'delete')
class FirmwareAssignmentDeleteView(generic.ObjectDeleteView):
    queryset = models.FirmwareAssignment.objects.all()
    default_return_url = 'plugins:netbox_firmware:firmwareassignment_list'
    
    def post(self, request, *args, **kwargs):
        return super().post(request, *args, **kwargs)

# ----------------- Bulk Import, Edit, Delete -----------------

@register_model_view(models.FirmwareAssignment, 'bulk_import', detail=False)
class FirmwareBulkImportView(generic.BulkImportView):
    queryset = models.FirmwareAssignment.objects.all()
    model_form = forms.FirmwareAssignmentImportForm

    def save_object(self, object_form, request):
        obj = object_form.save()
        return obj

@register_model_view(models.FirmwareAssignment, 'bulk_edit', detail=False)
class FirmwareAssignmentBulkEditView(generic.BulkEditView):
    queryset = models.FirmwareAssignment.objects.all()
    filterset = filtersets.FirmwareAssignmentFilterSet
    table = tables.FirmwareAssignmentTable
    form = forms.FirmwareAssignmentBulkEditForm
    default_return_url = 'plugins:netbox_firmware:firmwareassignment_list'
    
    def post (self, request, **kwargs):
        return super().post(request, **kwargs)

@register_model_view(models.FirmwareAssignment, 'bulk_delete', detail=False)
class FirmwareAssignmentBulkDeleteView(generic.BulkDeleteView):
    queryset = models.FirmwareAssignment.objects.all()
    table = tables.FirmwareAssignmentTable

    def post(self, request):
        return super().post(request)


# Generated by Django 5.2.1 on 2025-09-19 12:03

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dcim', '0207_remove_redundant_indexes'),
        ('netbox_firmware', '0004_firmware_manufacturer_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='firmware',
            name='manufacturer',
            field=models.ForeignKey(default=7, on_delete=django.db.models.deletion.PROTECT, related_name='Firmware', to='dcim.manufacturer'),
        ),
    ]



from django.db import models
from django.forms import ValidationError
from django.urls import reverse
from django.db.models.functions import Lower
from django.utils.translation import gettext_lazy as _

from ..choices import HardwareKindChoices, BiosStatusChoices
from netbox.models import NetBoxModel, ChangeLoggedModel, NestedGroupModel
from dcim.models import Manufacturer, DeviceType, ModuleType, Device, Module


### ------------------------------------------------------- ###
### BIOS Model
### ------------------------------------------------------- ###


class Bios(NetBoxModel):
    #
    # fields that identify bios
    #
    name = models.CharField(
        help_text='Name of the bios',
        max_length=255,
        verbose_name='Name',
    )
    manufacturer = models.ForeignKey(
        help_text='Name of the manufacturer',
        to= 'dcim.Manufacturer',
        on_delete=models.PROTECT,
        related_name="bios",
        # default=22,  # Uncomment and set to a valid Manufacturer ID for existing records without a manufacturer.
    )
    file_name = models.CharField(
        help_text='File name of the bios',
        blank=True,
        null=True,
        max_length=255,
        verbose_name='File Name',
    )
    file = models.FileField(
        upload_to='bios-files',
        help_text='File of the bios',
        blank=True,
        null=True,
        verbose_name='File',
    )
    status = models.CharField(
        max_length=50,
        choices= BiosStatusChoices,
        default= BiosStatusChoices.STATUS_ACTIVE,
        help_text='Bios lifecycle status',
    )
    description = models.CharField(
        help_text='Description of the bios',
        max_length=255,
        verbose_name='Description',
        null=True,
        blank=True
    )
    comments = models.TextField(
        blank=True,
        null=True,
        help_text='Additional comments about the bios',
    )
    
    #
    ### Hardware Type fields ###
    #

    device_type = models.ManyToManyField(
        to=DeviceType,
        related_name='bios',
        blank=True,
        verbose_name='Device Type',
    )
    module_type = models.ManyToManyField(
        to=ModuleType,
        related_name='bios',
        blank=True,
        verbose_name='Module Type',
    )
    
    clone_fields = [
        'name', 'manufacturer', 'description', 'status', 'device_type', 'module_type'
    ]
    

    class Meta:
        ordering = ('name',)
        verbose_name = 'BIOS'
        verbose_name_plural = 'BIOS'
        constraints = [
            models.UniqueConstraint(
                Lower('name'),
                name='%(app_label)s_%(class)s_unique_name',
                violation_error_message=_("The BIOS 'Name' must be unique.")
            ),
        ]
    
    ### Get all fields of the model ?
    @classmethod
    def get_fields(cls):
        return {field.name: field for field in cls._meta.get_fields()}

        #
    ### Properties ###
    #

    ### This property is needed to check if the BIOS is for Devices or Modules and also to filter the template and show only the assigned Device Types / Hide Module Types.
    ### We need to use .exists() because it's a ManyToMany field.
    @property
    def kind(self):
        if self.device_type_id:
            return HardwareKindChoices.DEVICE
        elif self.module_type_id:
            return HardwareKindChoices.MODULE
        else:
            return None
           
    ### Needed to show File Path in the Firmware Table
    @property
    def file_path(self):
        return self.file.name if self.file else None

    @property
    def hardware_type(self):
        return self.device_type or self.module_type or None

    ### This property is needed to filter the template and show only the assigned Device Types / Hide Module Types
    @property
    def has_device_type(self):
        return self.device_type.exists()

    ### This property is needed to filter the template and show only the assigned Module Types / Hide Device Types
    @property
    def has_module_type(self):
        return self.module_type.exists()
    
    #
    ### Main definitions ###
    #

    ### This is needed for bulk import/export and other places where the kind of hardware is needed.    
    def get_kind_display(self):
        return dict(HardwareKindChoices)[self.kind]

    ### This will disable changing the Manufacturer once a Device Type or Module Type is assigned
    def clean(self):
        if self.pk:
            old = Bios.objects.get(pk=self.pk)

            # Check if manufacturer changed
            if old.manufacturer != self.manufacturer:
                # Only lock if already linked to device_type(s) or module_type(s)
                if self.device_type.exists() or self.module_type.exists():
                    raise ValidationError({
                        "manufacturer": (
                            "The manufacturer cannot be changed because this bios "
                            "is already linked to one or more device types or module types."
                        )
                    })

        return super().clean()

    ### This will return the URL to the bios detail view.
    def get_absolute_url(self):
        return reverse('plugins:netbox_firmware:bios', args=[self.pk])

    ### This will delete the file from the storage when the BIOS object is deleted.
    def delete(self,*args, **kwargs):
        _name = self.file.name
        super().delete(*args, **kwargs)
        self.file.delete(save=False)
        self.file.name = _name

    ### This is how the firmware will be displayed in the lists using his name.
    def __str__(self):
        return f'{self.name}'



### ------------------------------------------------------ ###
### BIOS Assignment Model
### ------------------------------------------------------ ###


class BiosAssignment(NetBoxModel):
    description = models.TextField(blank=True, null=True)
    ticket_number = models.CharField(max_length=100, blank=True, null=True)
    patch_date = models.DateField(blank=True, null=True)
    comment = models.TextField(blank=True, null=True)
    bios = models.ForeignKey(
        to=Bios,
        related_name='BiosAssignment',
        on_delete=models.PROTECT,
        verbose_name='BIOS',
        null=True,
        blank=True
    )
    module = models.ForeignKey(
        to=Module,
        related_name='BiosAssignment',
        on_delete=models.PROTECT,
        verbose_name='Module',
        null=True,
        blank=True
    )
    device = models.ForeignKey(
        to=Device, 
        related_name='BiosAssignment',
        on_delete=models.PROTECT,
        verbose_name='Device',
        null=True, 
        blank=True
    )

    clone_fields = [
        'bios', 'patch_date', 'description', 'comment', 'module', 'device'
    ]

    class Meta:
        ordering = ('bios', 'device', 'module')
        verbose_name = 'BIOS Assignment'
        verbose_name_plural = 'BIOS Assignments'
        constraints = [
            models.CheckConstraint(
                check=models.Q(device__isnull=False) | models.Q(module__isnull=False) ,
                name='bios_device_or_module_required'
            ),
            models.UniqueConstraint(fields=['device'], name='unique_bios_per_device'),
            models.UniqueConstraint(fields=['module'], name='unique_bios_per_module'),
        ]

    @property
    def kind(self):
        if self.device_id:
            return HardwareKindChoices.DEVICE
        elif self.module_id:
            return HardwareKindChoices.MODULE
        else:
            return None
        
    def get_kind_display(self):
        return dict(HardwareKindChoices)[self.kind]

    @property
    def module_device(self):
        return self.module.device if self.module else None
    
    @property
    def hardware_sn(self):
        return self.device.serial if self.device else self.module.serial if self.module else None

    @property
    def hardware_type(self):
        if self.device and self.device.device_type:
            return self.device.device_type
        if self.module and self.module.module_type:
            return self.module.module_type
        return None
    
    @property
    def hardware(self):
        return self.device or self.module or None
    
    @property
    def device_sn(self):
        return self.device.serial if self.device else None
    
    @property
    def module_sn(self):
        return self.module.serial if self.module else None
    
    @property
    def device_type(self):
        return self.device.device_type if self.device else None
    
    @property
    def module_type(self):
        return self.module.module_type if self.module else None

    @property
    def manufacturer(self):
        return self.get_manufacturer()

    def get_manufacturer(self):
        """ Haalt de manufacturer op afhankelijk van device_type of module_type """
        if self.device:
            return self.device.device_type.manufacturer  # Haal de fabrikant via device_type
        elif self.module:
            return self.module.module_type.manufacturer  # Haal de fabrikant via module_type
        print('No manufacturer found')
        return None

    def __str__(self):
        if self.hardware:
            return f"{self.bios} - {self.hardware}"
        return f"{self.bios}"




from netbox.api.serializers import NetBoxModelSerializer
from dcim.api.serializers import ManufacturerSerializer, DeviceTypeSerializer, ModuleTypeSerializer, DeviceSerializer, ModuleSerializer
from netbox_firmware.models import Firmware, FirmwareAssignment

__all__ = (
    'FirmwareSerializer',
    'FirmwareAssignmentSerializer',
)

class FirmwareSerializer(NetBoxModelSerializer):
    manufacturer = ManufacturerSerializer(nested=True, required=True)
    device_type = DeviceTypeSerializer(many=True, nested=True, required=False)
    module_type = ModuleTypeSerializer(many=True, nested=True, required=False)
    class Meta:
        model = Firmware
        fields = '__all__'


class FirmwareAssignmentSerializer(NetBoxModelSerializer):
    firmware = FirmwareSerializer(nested=True, required=True)
    device = DeviceSerializer(nested=True, required=False)
    module = ModuleSerializer(nested=True, required=False)
    class Meta:
        model = FirmwareAssignment
        fields = '__all__'


from django.db import models
from django.forms import ValidationError
from django.urls import reverse
from django.db.models.functions import Lower
from django.utils.translation import gettext_lazy as _

from ..choices import HardwareKindChoices, FirmwareStatusChoices
from netbox.models import NetBoxModel, ChangeLoggedModel, NestedGroupModel
from dcim.models import Manufacturer, DeviceType, ModuleType, Device, Module


### ------------------------------------------------------- ###
### Firmware Model
### ------------------------------------------------------- ###


class Firmware(NetBoxModel):
    #
    ### Fields that identify firmware ###
    #
    name = models.CharField(
        help_text='Name of the firmware',
        max_length=255,
        verbose_name='Name',
    )
    manufacturer = models.ForeignKey(
        help_text='Name of the manufacturer',
        to= 'dcim.Manufacturer',
        on_delete=models.PROTECT,
        related_name="Firmware",
        # default=22,  # Uncomment and set to a valid Manufacturer ID for existing records without a manufacturer.
    )
    file_name = models.CharField(
        help_text='File name of the firmware',
        blank=True,
        null=True,
        max_length=255,
        verbose_name='File Name',
    )
    file = models.FileField(
        upload_to='firmware-files',
        help_text='File of the firmware',
        blank=True,
        null=True,
        verbose_name='File',
    )
    status = models.CharField(
        max_length=50,
        choices= FirmwareStatusChoices,
        default= FirmwareStatusChoices.STATUS_ACTIVE,
        help_text='Firmware lifecycle status',
    )
    description = models.CharField(
        help_text='Description of the firmware',
        max_length=255,
        verbose_name='Description',
        null=True,
        blank=True
    )
    comments = models.TextField(
        blank=True,
        null=True,
        help_text='Additional comments about the firmware',
    )
    
    #
    ### Hardware Type fields ###
    #

    device_type = models.ManyToManyField(
        to=DeviceType,
        related_name='firmware',
        blank=True,
        verbose_name='Device Type',
    )
    module_type = models.ManyToManyField(
        to=ModuleType,
        related_name='firmware',
        blank=True,
        verbose_name='Module Type',
    )
    
    clone_fields = [
        'name', 'manufacturer', 'description', 'status', 'device_type', 'module_type'
    ]


    class Meta:
        ordering = ('name',)
        verbose_name = 'Firmware'
        verbose_name_plural = 'Firmwares'
        constraints = [
            models.UniqueConstraint(
                Lower('name'),
                name='%(app_label)s_%(class)s_unique_name',
                violation_error_message=_("The Firmware 'Name' must be unique.")
            ),
        ]

    ### Get all fields of the model ?
    @classmethod
    def get_fields(cls):
        return {field.name: field for field in cls._meta.get_fields()}

    #
    ### Properties ###
    #

    ### This property is needed to check if the firmware is for Devices or Modules and also to filter the template and show only the assigned Device Types / Hide Module Types.
    ### We need to use .exists() because it's a ManyToMany field.
    @property
    def kind(self):
        if self.device_type.exists():
            return HardwareKindChoices.DEVICE
        elif self.module_type.exists():
            return HardwareKindChoices.MODULE
        return None

    ### Needed to show File Path in the Firmware Table
    @property
    def file_path(self):
        return self.file.name if self.file else None

    @property
    def hardware_type(self):
        return self.device_type or self.module_type or None

    ### This property is needed to filter the template and show only the assigned Device Types / Hide Module Types
    @property
    def has_device_type(self):
        return self.device_type.exists()

    ### This property is needed to filter the template and show only the assigned Module Types / Hide Device Types
    @property
    def has_module_type(self):
        return self.module_type.exists()


    #
    ### Main definitions ###
    #

    ### This is needed for bulk import/export and other places where the kind of hardware is needed.    
    def get_kind_display(self):
        return dict(HardwareKindChoices)[self.kind]

    ### This will disable changing the Manufacturer once a Device Type or Module Type is assigned
    def clean(self):
        if self.pk:
            old = Firmware.objects.get(pk=self.pk)

            # Check if manufacturer changed
            if old.manufacturer != self.manufacturer:
                # Only lock if already linked to device_type(s) or module_type(s)
                if self.device_type.exists() or self.module_type.exists():
                    raise ValidationError({
                        "manufacturer": (
                            "The manufacturer cannot be changed because this firmware "
                            "is already linked to one or more device types or module types."
                        )
                    })

        return super().clean()

    ### This will return the URL to the firmware detail view.
    def get_absolute_url(self):
        return reverse('plugins:netbox_firmware:firmware', args=[self.pk])
    
    ### This will delete the file from the storage when the Firmware object is deleted.
    def delete(self,*args, **kwargs):
        _name = self.file.name
        super().delete(*args, **kwargs)
        self.file.delete(save=False)
        self.file.name = _name

    ### This is how the firmware will be displayed in the lists using his name.
    def __str__(self):
        return f'{self.name}'



### ------------------------------------------------------ ###
### Firmware Assignment Model
### ------------------------------------------------------ ###


class FirmwareAssignment(NetBoxModel):
    description = models.TextField(blank=True, null=True)
    ticket_number = models.CharField(max_length=100, blank=True, null=True)
    patch_date = models.DateField(blank=True, null=True)
    comment = models.TextField(blank=True, null=True)
    firmware = models.ForeignKey(
        to=Firmware,
        related_name='FirmwareAssignment',
        on_delete=models.PROTECT,
        verbose_name='Firmware',
        null=True,
        blank=True
    )
    module = models.ForeignKey(
        to=Module,
        related_name='FirmwareAssignment',
        on_delete=models.PROTECT,
        verbose_name='Module',
        null=True,
        blank=True
    )
    device = models.ForeignKey(
        to=Device, 
        related_name='FirmwareAssignment',
        on_delete=models.PROTECT,
        verbose_name='Device',
        null=True, 
        blank=True
    )

    clone_fields = [
        'firmware', 'patch_date', 'description', 'comment', 'module', 'device'
    ]

    class Meta:
        ordering = ('firmware', 'device', 'module')
        verbose_name = 'Firmware Assignment'
        verbose_name_plural = 'Firmware Assignments'
        constraints = [
            models.CheckConstraint(
                check=models.Q(device__isnull=False) | models.Q(module__isnull=False),
                name='firmassign_either_device_or_module_required'
            ),
            models.UniqueConstraint(fields=['device'], name='unique_firmware_per_device'),
            models.UniqueConstraint(fields=['module'], name='unique_firmware_per_module'),
        ]

    #
    ### Properties ###
    #

    @property
    def kind(self):
        if self.device_id:
            return HardwareKindChoices.DEVICE
        elif self.module_id:
            return HardwareKindChoices.MODULE
        else:
            return None
        
    @property
    def hardware(self):
        return self.device or self.module or None

    @property
    def hardware_type(self):
        return self.device_type or self.module_type or None

    @property
    def hardware_sn(self):
        return self.device.serial if self.device else self.module.serial if self.module else None

    @property
    def device_sn(self):
        return self.device.serial if self.device else None
    
    @property
    def module_device(self):
        return self.module.device if self.module else None
    
    @property
    def module_sn(self):
        return self.module.serial if self.module else None

    @property
    def device_type(self):
        return self.device.device_type if self.device else None
    
    @property
    def module_type(self):
        return self.module.module_type if self.module else None

    @property
    def manufacturer(self):
        return self.get_manufacturer()

    #
    ### Main definitions ###
    #

    def get_kind_display(self):
        return dict(HardwareKindChoices)[self.kind]

    ### Needed to get the manufacturer depending on device_type or module_type
    def get_manufacturer(self):
        """ Get manufacturer depending on device_type or module_type """
        if self.device:
            return self.device.device_type.manufacturer
        elif self.module:
            return self.module.module_type.manufacturer
        print('No manufacturer found')
        return None

    def __str__(self):
        if self.hardware:
            return f"{self.firmware} - {self.hardware}"
        return f"{self.firmware}"


from .schema import (
    FirmwareQuery,
    FirmwareAssignmentQuery,
    BiosQuery,
    BiosAssignmentQuery,
)

schema = [
    FirmwareQuery,
    FirmwareAssignmentQuery,
    BiosQuery,
    BiosAssignmentQuery,
]


import django_filters
from django.db.models import Q
from utilities.filters import (
    ContentTypeFilter, MultiValueCharFilter, MultiValueMACAddressFilter, MultiValueNumberFilter, MultiValueWWNFilter,
    NumericArrayFilter, TreeNodeMultipleChoiceFilter,
)
from django.utils.translation import gettext as _
from dcim.models import DeviceType, Manufacturer, ModuleType, Device, Module
from netbox_firmware.choices import FirmwareStatusChoices, HardwareKindChoices

from netbox.filtersets import NetBoxModelFilterSet
from ..models import Firmware, FirmwareAssignment

class FirmwareFilterSet(NetBoxModelFilterSet):
    name = MultiValueCharFilter(
        lookup_expr='iexact',
    )
    file_name = MultiValueCharFilter(
        lookup_expr='icontains',
        label=_('File name'),
    )
    status = django_filters.MultipleChoiceFilter(
        choices=FirmwareStatusChoices,
        label=_('Status'),
    )
    kind = MultiValueCharFilter(
        method='filter_kind',
        label='Type of hardware',
    )
    manufacturer = django_filters.ModelMultipleChoiceFilter(
        field_name='manufacturer__name',
        queryset=Manufacturer.objects.all(),
        to_field_name='name',
        label=_('Manufacturer (Name)'),
    )
    manufacturer_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Manufacturer.objects.all(),
        label=_('Manufacturer (ID)'),
    )
    device_type = django_filters.ModelMultipleChoiceFilter(
        field_name='device_type__model',
        queryset=DeviceType.objects.all(),
        to_field_name='model',
        label=_('Device type (Model)'),
    )
    device_type_id = django_filters.ModelMultipleChoiceFilter(
        field_name='device_type',
        queryset=DeviceType.objects.all(),
        label=_('Device type (ID)'),
    )
    device = django_filters.ModelChoiceFilter(
        queryset=Device.objects.all(),
        method='filter_by_device',
        label=_('Device'),
    )
    module_type = django_filters.ModelMultipleChoiceFilter(
        field_name='module_type__model',
        queryset=ModuleType.objects.all(),
        to_field_name='model',
        label=_('Module type (Model)'),
    )
    module_type_id = django_filters.ModelMultipleChoiceFilter(
        field_name='module_type',
        queryset=ModuleType.objects.all(),
        label=_('Module type (ID)'),
    )
    module = django_filters.ModelChoiceFilter(
        queryset=Module.objects.all(),
        method='filter_by_module',
        label=_('Module'),
    )
    
    ## In the class Meta you can only add field names that are actual fields of the model.
    class Meta:
        model = Firmware
        fields = {
            'id', 'name', 'file_name', 'status',
            'device_type', 'module_type', 
        }
    
    ### Criteria for the Quick search box.
    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(name__icontains=value) |
            Q(manufacturer__name__icontains=value) |
            Q(device_type__model__icontains=value) |
            Q(module_type__model__icontains=value) 
        ).distinct()
    
    def filter_kind(self, queryset, name, value):
        query = None
        for kind in HardwareKindChoices.values():
            if kind in value:
                q = Q(**{f'{kind}_type__isnull': False})
                if query:
                    query = query | q
                else:
                    query = q
        if query:
            return queryset.filter(query).distinct()
        else:
            return queryset

    ### Filter needed to show only the compatible firmwares for a device during assignment
    def filter_by_device(self, queryset, name, value):
        return queryset.filter(device_type=value.device_type)

    ### Filter needed to show only the compatible firmwares for a module during assignment
    def filter_by_module(self, queryset, name, value):
        return queryset.filter(module_type=value.module_type)


class FirmwareAssignmentFilterSet(NetBoxModelFilterSet):
    description = MultiValueCharFilter(
        lookup_expr='icontains',
    )
    ticket_number = MultiValueCharFilter(
        lookup_expr='icontains',
        label=_('Ticket number'),
    )
    patch_date = django_filters.DateFromToRangeFilter(
        label=_('Patch date'),
    )
    comment = MultiValueCharFilter(
        lookup_expr='icontains',
    )
    firmware = django_filters.ModelMultipleChoiceFilter(
        field_name='firmware__name',
        queryset=Firmware.objects.all(),
        to_field_name='name',
        label=_('Firmware'),
    )
    firmware_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Firmware.objects.all(),
        label=_('Firmware (ID)'),
    )
    kind = MultiValueCharFilter(
        method='filter_kind',
        label='Type of hardware',
    )
    manufacturer_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Manufacturer.objects.all(),
        label=_('Manufacturer (ID)'),
        method='filter_by_manufacturer',  # Custom filter method
    )
    device_type = django_filters.ModelMultipleChoiceFilter(
        queryset=DeviceType.objects.all(),
        field_name='device__device_type__model',
        to_field_name='model',
        label=_('Device type (model)'),
    )
    device_type_id = django_filters.ModelMultipleChoiceFilter(
        queryset=DeviceType.objects.all(),
        field_name='device__device_type',
        label=_('Device type (ID)'),
    )
    device = django_filters.ModelMultipleChoiceFilter(
        queryset=Device.objects.all(),
        field_name='device__name',
        to_field_name='name',
        label=_('Device (name)'),
    )
    device_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Device.objects.all(),
        label=_('Device (ID)'),
    )
    device_sn = django_filters.ModelMultipleChoiceFilter(
        queryset=Device.objects.all(),
        field_name='device__serial',
        label=_('Device Serial Number'),
    )
    module_type = django_filters.ModelMultipleChoiceFilter(
        queryset=ModuleType.objects.all(),
        field_name='module__module_type__model',
        to_field_name='model',
        label=_('Module type (model)'),
    )
    module_type_id = django_filters.ModelMultipleChoiceFilter(
        queryset=ModuleType.objects.all(),
        field_name='module__module_type',
        label=_('Module type (ID)'),
    )
    module_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Module.objects.all(),
        label=_('Module (ID)'),
    )
    module_sn = django_filters.CharFilter(
        field_name='module__serial',
        lookup_expr='icontains',
        label=_('Module Serial Number'),
    )
    module_device = django_filters.ModelMultipleChoiceFilter(
        queryset=Module.objects.all(),
        field_name='module__device',
        label=_('Module (device)'),
    )
    module_device_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Device.objects.all(),
        field_name='module__device_id',
        label=_('Module (device ID)'),
    )



    ## In the class Meta you can only add field names that are actual fields of the model.
    class Meta:
        model = FirmwareAssignment
        fields = {
            'id', 'description', 'ticket_number', 'patch_date', 'comment',
            'firmware', 'device', 'module',
        }
    
    ### Criteria for the Quick search box.
    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(description__icontains=value) |
            Q(ticket_number__icontains=value) |
            Q(comment__icontains=value) |
            Q(firmware__name__icontains=value) |
            Q(device__device_type__model__icontains=value) |
            Q(module__module_type__model__icontains=value) |
            Q(device__name__icontains=value) |
            Q(device__serial__icontains=value) |
            Q(module__serial__icontains=value) |
            Q(device__device_type__manufacturer__name__icontains=value) |
            Q(module__module_type__manufacturer__name__icontains=value) 
        ).distinct()

    def filter_kind(self, queryset, name, value):
        query = None
        for kind in HardwareKindChoices.values():
            if kind in value:
                q = Q(**{f'{kind}__isnull': False})
                if query:
                    query = query | q
                else:
                    query = q
        if query:
            return queryset.filter(query)
        else:
            return queryset

    ### Custom filter to filter by manufacturer, checking both device and module manufacturer
    def filter_by_manufacturer(self, queryset, name, value):
        if value:
            return queryset.filter(
                Q(device__device_type__manufacturer__in=value) |
                Q(module__module_type__manufacturer__in=value)
            ).distinct()
        return queryset


from django import forms
from django.utils.translation import gettext_lazy as _

from dcim.models import DeviceType, Manufacturer, ModuleType, Device, Module
from netbox.forms import NetBoxModelBulkEditForm
from utilities.forms.fields import (
    CommentField,
    DynamicModelChoiceField,
    DynamicModelMultipleChoiceField,
)
from utilities.forms.widgets import DatePicker
from utilities.forms.rendering import FieldSet, TabbedGroups
from utilities.forms import add_blank_choice

from netbox_firmware.choices import BiosStatusChoices
from netbox_firmware.models import (
    Bios,
    BiosAssignment
)

class BiosBulkEditForm(NetBoxModelBulkEditForm):
    name = forms.CharField(required=False, label='Name')
    status = forms.ChoiceField(
        choices=add_blank_choice(BiosStatusChoices),
        required=False,
        label='Status',
    )
    description = forms.CharField(
        required=False,
    )
    file_name = forms.CharField(required=False, label='File Name')
    manufacturer = DynamicModelChoiceField(
        queryset=Manufacturer.objects.all(),
        required=False,
        selector=True,
        label='Manufacturer'
    )
    device_type = DynamicModelMultipleChoiceField(
        queryset=DeviceType.objects.all(),
        required=False,
        selector=True,
        label='Supported Device Type',
        query_params={
            'manufacturer_id': '$manufacturer',
        },
    )
    add_device_type = DynamicModelMultipleChoiceField(
        label=_('Add Device Type'),
        queryset=DeviceType.objects.all(),
        required=False,
        query_params={
            'manufacturer_id': '$manufacturer',
        },
    )
    remove_device_type = DynamicModelMultipleChoiceField(
        label=_('Remove Device Type'),
        queryset=DeviceType.objects.all(),
        required=False,
        query_params={
            'manufacturer_id': '$manufacturer',
        }
    )
    module_type = DynamicModelMultipleChoiceField(
        queryset=ModuleType.objects.all(),
        required=False,
        selector=True,
        label='Supported Module Type',
        query_params={
            'manufacturer_id': '$manufacturer',
        },
    )
    add_module_type = DynamicModelMultipleChoiceField(
        label=_('Add Module Type'),
        queryset=ModuleType.objects.all(),
        required=False,
        query_params={
            'manufacturer_id': '$manufacturer',
        },
    )
    remove_module_type = DynamicModelMultipleChoiceField(
        label=_('Remove Module Type'),
        queryset=ModuleType.objects.all(),
        required=False,
        query_params={
            'manufacturer_id': '$manufacturer',
        }
    )
    comments = CommentField()
    


    model = Bios
    fieldsets=(
        FieldSet('name', 'manufacturer', 'file_name', 'status', 'description',name='General'),
        FieldSet(
            TabbedGroups(
                FieldSet('device_type', name=_('Fixed Assignment')),
                FieldSet('add_device_type', 'remove_device_type', name=_('Add/Remove')),
            ),
            name='Device Type'
        ),
        FieldSet(
            TabbedGroups(
                FieldSet('module_type', name=_('Fixed Assignment')),
                FieldSet('add_module_type', 'remove_module_type', name=_('Add/Remove')),
            ),
            name='Module Type'
        ),
    )
    nullable_fields = ['device_type', 'module_type']
    

class BiosAssignmentBulkEditForm(NetBoxModelBulkEditForm):
    description = forms.CharField(
        required=False,
    )
    
    # Hardware Items ------------------------
    
    device = DynamicModelChoiceField(
        queryset = Device.objects.all(),
        required=False,
        selector=True,
        label='Device',
    )
    module = DynamicModelChoiceField(
        queryset = Module.objects.all(),
        required=False,
        selector=True,
        label='Module',
    )
    
    # Update --------------------------------
    bios = DynamicModelChoiceField(
        queryset=Bios.objects.all(),
        selector=True,
        required=False,
        label='Bios',
    )
    comment = CommentField()
    patch_date = forms.DateField(
        widget=DatePicker(attrs={'is_clearable': True}),
        required=False,
        label='Patch Date',
        help_text='Date of the bios patch'
    )
    ticket_number = forms.CharField(
        required=False,
        label='Ticket Number',
        help_text='Ticket number for the bios patch'
    )
    
    model= BiosAssignment
    fieldsets = (
        FieldSet(
            'description',
            TabbedGroups(
                FieldSet('device',name='Device'),
                FieldSet('module',name='Module'),
            ),
            name='Hardware'
        ),
        FieldSet(
            'ticket_number','bios','patch_date','comment',
            name='Update'
        ),
    )
    nullable_fields = ['device', 'module', 'ticket_number', 'patch_date', 'comment']


from utilities.choices import ChoiceSet
from django.utils.translation import gettext_lazy as _


#
# Firmware
#

class FirmwareStatusChoices(ChoiceSet):
    key = 'Firmware.status'

    STATUS_ACTIVE = 'active'
    STATUS_STAGED = 'staged'
    STATUS_PLANNED = 'planned'
    STATUS_DECOMMISSIONING = 'decommissioning'
    STATUS_CORRUPTED = 'corrupted'
    STATUS_ARCHIVED = 'archived'
    

    CHOICES = [
        (STATUS_ACTIVE, _('Active'), 'green'),
        (STATUS_STAGED, _('Staged'), 'blue'),
        (STATUS_PLANNED, _('Planned'), 'cyan'),
        (STATUS_DECOMMISSIONING, _('Decommissioning'), 'yellow'),
        (STATUS_CORRUPTED, _('Corrupted'), 'red'),
        (STATUS_ARCHIVED, _('Archived'), 'purple'),
        
    ]

#
# Bios
#

class BiosStatusChoices(ChoiceSet):
    key = 'Bios.status'

    STATUS_ACTIVE = 'active'
    STATUS_STAGED = 'staged'
    STATUS_PLANNED = 'planned'
    STATUS_DECOMMISSIONING = 'decommissioning'
    STATUS_CORRUPTED = 'corrupted'
    STATUS_ARCHIVED = 'archived'
    

    CHOICES = [
        (STATUS_ACTIVE, _('Active'), 'green'),
        (STATUS_STAGED, _('Staged'), 'blue'),
        (STATUS_PLANNED, _('Planned'), 'cyan'),
        (STATUS_DECOMMISSIONING, _('Decommissioning'), 'yellow'),
        (STATUS_CORRUPTED, _('Corrupted'), 'red'),
        (STATUS_ARCHIVED, _('Archived'), 'purple'),
        
    ]

#
# General
#


class HardwareKindChoices(ChoiceSet):
    DEVICE = 'device'
    MODULE = 'module'

    CHOICES = [
        (DEVICE, 'Device'),
        (MODULE, 'Module'),
    ]


from dcim.models import DeviceType, Manufacturer, ModuleType, Device, Module
from django import forms
from django.core.exceptions import ValidationError
from netbox.forms import NetBoxModelForm
from utilities.forms.fields import CommentField, DynamicModelChoiceField, DynamicModelMultipleChoiceField
from utilities.forms.rendering import FieldSet, TabbedGroups
from utilities.forms.widgets import DatePicker, ClearableFileInput
from netbox_firmware.utils import get_tags_and_edit_protected_firmware_fields
from netbox_firmware.filtersets import FirmwareFilterSet, FirmwareAssignmentFilterSet
from netbox_firmware.models import Firmware, FirmwareAssignment

__all__ = (
    'FirmwareForm',
    'FirmwareAssignmentForm',
)

class FirmwareForm(NetBoxModelForm):
    name = forms.CharField()
    manufacturer = DynamicModelChoiceField(
        queryset=Manufacturer.objects.all(),
        required=True,
        label="Manufacturer",
        selector=True,
        quick_add=True
    )
    description = forms.CharField(
        required=False,
    )
    file_name = forms.CharField(required=False, label='File Name')
    device_type = DynamicModelMultipleChoiceField(
        queryset=DeviceType.objects.all(),
        query_params={
            'manufacturer_id': '$manufacturer',
        },
        required=False,
        selector=True,
        label='Supported Device Type',
    )
    module_type = DynamicModelMultipleChoiceField(
        queryset=ModuleType.objects.all(),
        query_params={
            'manufacturer_id': '$manufacturer',
        },
        required=False,
        selector=True,
        label='Supported Module Type',
    )
    comments = CommentField()
    
    fieldsets=(
        FieldSet('name', 'manufacturer', 'file_name', 'file', 'status', 'description',name='General'),
        FieldSet(
            TabbedGroups(
                FieldSet('device_type',name='Device Type'),
                FieldSet('module_type',name='Module Type')
            ),
            name='Hardware'
        ),
    )

    class Meta:
        model = Firmware
        fields = '__all__'
        widgets = {
            'file': ClearableFileInput(attrs={
                'accept': '.bin,.img,.tar,.tar.gz,.zip,.exe'
                }),
        }


    ### This checks what type is assigned to the firmware and will filter the template to show only one the type used.
    ### IT will also grey out the Manufacturer field once a Device Type or Module Type is assigned.
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Default: assume no hardware type selected
        self.no_hardware_type = True

        if self.instance and self.instance.pk:
            has_device_type = self.instance.device_type.exists()
            has_module_type = self.instance.module_type.exists()

            if has_device_type or has_module_type:
                # Used for template logic (which tab to show)
                self.no_hardware_type = False

                # Disable manufacturer field once firmware is linked
                self.fields['manufacturer'].disabled = True

            # Disable device_type if module_type is selected
            if has_module_type:
                self.fields['device_type'].disabled = True

            # Disable module_type if device_type is selected
            if has_device_type:
                self.fields['module_type'].disabled = True


    ### This ensures that only one of Device Type or Module Type can be selected during firmware creation.
    def clean(self):
        super().clean()
        device_types = self.cleaned_data.get('device_type')
        module_types = self.cleaned_data.get('module_type')

        if device_types and module_types:
            raise forms.ValidationError("You may only select one of 'Device Type' or 'Module Type', not both.")

        return self.cleaned_data
      
class FirmwareAssignmentForm(NetBoxModelForm):
    # Hardware ------------------------------
    description = forms.CharField(
        required=False,
    )

    # Hardware Items ------------------------
    device = DynamicModelChoiceField(
        queryset = Device.objects.all(),
        required=False,
        selector=True,
        label='Device',
        query_params={
            'manufacturer_id': '$manufacturer',
        },
    )
    module = DynamicModelChoiceField(
        queryset = Module.objects.all(),
        required=False,
        selector=True,
        label='Module',
        query_params={
            'manufacturer_id': '$manufacturer',
        },
    )
    
    # Update --------------------------------
    firmware = DynamicModelChoiceField(
        queryset=Firmware.objects.all(),
        selector=True,
        required=True,
        label='Firmware',
        help_text='Only showing Active and Staged',
        query_params={
            'status__in': ['active','staged'],
            'device': '$device',
            'module': '$module',
        },
    )
    comment = CommentField()
    
    fieldsets = (
        FieldSet(
            'description',
            TabbedGroups(
                FieldSet('device',name='Device'),
                FieldSet('module',name='Module'),
            ),
            name='Hardware'
        ),
        FieldSet(
            'ticket_number','firmware','patch_date','comment',
            name='Update'
        ),
    )
    
    class Meta:
        model = FirmwareAssignment
        fields = [
            'description',
            'ticket_number',
            'patch_date',
            'comment',
            'firmware',
            'device',
            'module',
        ]
        widgets = {
            'patch_date': DatePicker(),
        }
    
    def clean(self):
        super().clean()
        device = self.cleaned_data.get('device')
        module = self.cleaned_data.get('module')

        if device and module:
            raise forms.ValidationError("You may only select one of 'Device' or 'Module', not both.")
        
        pass


import logging
from django.contrib import messages
from django.db import IntegrityError
from django.shortcuts import redirect
from django.template import Template
from netbox.views import generic
from utilities.query import count_related
from utilities.views import register_model_view, ViewTab
from ..utils import get_countdevice

from netbox_firmware import tables
from netbox_firmware import forms
from netbox_firmware import models
from netbox_firmware import filtersets

__all__ = (
    'BiosView',
    'BiosListView'
)

@register_model_view(models.Bios)
class BiosView(generic.ObjectView):
    queryset = models.Bios.objects.all()

    def get_extra_context(self, request, instance):
        context = super().get_extra_context(request, instance)
        return context

@register_model_view(models.Bios, 'list', path='', detail=False)
class BiosListView(generic.ObjectListView):
    queryset = models.Bios.objects.prefetch_related(
        'device_type',
        'module_type'
    ).annotate(
        instance_count=count_related(models.BiosAssignment,'bios'),
    )
    filterset = filtersets.BiosFilterSet
    filterset_form = forms.BiosFilterForm
    table = tables.BiosTable

@register_model_view(models.Bios, 'edit')
@register_model_view(models.Bios, 'add', detail=False)
class BiosEditView(generic.ObjectEditView):
    queryset = models.Bios.objects.all()
    form = forms.BiosForm
    default_return_url = 'plugins:netbox_firmware:bios_list'

@register_model_view(models.Bios,'delete')
class BiosDeleteView(generic.ObjectDeleteView):
    queryset = models.Bios.objects.all()
    default_return_url = 'plugins:netbox_firmware:bios_list'
    
    def post(self, request, *args, **kwargs):
        return super().post(request, *args, **kwargs)


# ----------------- Bulk Import, Edit, Delete -----------------

@register_model_view(models.Bios, 'bulk_import', detail=False)
class BiosBulkImportView(generic.BulkImportView):
    queryset = models.Bios.objects.all()
    model_form = forms.BiosImportForm

    def save_object(self, object_form, request):
        obj = object_form.save()
        return obj
    
@register_model_view(models.Bios, 'bulk_edit', detail=False)
class BiosBulkEditView(generic.BulkEditView):
    queryset = models.Bios.objects.all()
    filterset = filtersets.BiosFilterSet
    table = tables.BiosTable
    form = forms.BiosBulkEditForm
    default_return_url = 'plugins:netbox_firmware:bios_list'
    
    def post_save_operations(self, form, obj):
        super().post_save_operations(form, obj)

        # Add/remove Linked Device Types
        if form.cleaned_data.get('add_device_type', None):
            obj.device_type.add(*form.cleaned_data['add_device_type'])
        if form.cleaned_data.get('remove_device_type', None):
            obj.device_type.remove(*form.cleaned_data['remove_device_type'])

        # Add/remove Linked Module Types
        if form.cleaned_data.get('add_module_type', None):
            obj.module_type.add(*form.cleaned_data['add_module_type'])
        if form.cleaned_data.get('remove_module_type', None):
            obj.module_type.remove(*form.cleaned_data['remove_module_type'])

    def post (self, request, **kwargs):
        return super().post(request, **kwargs)

@register_model_view(models.Bios, 'bulk_delete', detail=False)
class BiosBulkDeleteView(generic.BulkDeleteView):
    queryset = models.Bios.objects.all()
    table = tables.BiosTable

    def post(self, request):
        return super().post(request)


# Generated by Django 5.2.1 on 2025-09-11 11:45

import django.db.models.deletion
import django.db.models.functions.text
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dcim', '0207_remove_redundant_indexes'),
        ('extras', '0128_tableconfig'),
        ('netbox_firmware', '0003_alter_firmware_options_alter_firmware_device_type_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='firmware',
            name='manufacturer',
            field=models.ForeignKey(default=6, on_delete=django.db.models.deletion.PROTECT, related_name='Firmware', to='dcim.manufacturer'),
        ),
        migrations.AlterConstraint(
            model_name='bios',
            name='netbox_firmware_bios_unique_name',
            constraint=models.UniqueConstraint(django.db.models.functions.text.Lower('name'), name='netbox_firmware_bios_unique_name', violation_error_message="The BIOS 'Name' must be unique."),
        ),
        migrations.AlterConstraint(
            model_name='firmware',
            name='netbox_firmware_firmware_unique_name',
            constraint=models.UniqueConstraint(django.db.models.functions.text.Lower('name'), name='netbox_firmware_firmware_unique_name', violation_error_message="The Firmware 'Name' must be unique."),
        ),
    ]



from netbox.plugins import PluginMenu, PluginMenuItem, PluginMenuButton, get_plugin_config

### Firmware ###

firmware_items = (
    PluginMenuItem(
        link='plugins:netbox_firmware:firmware_list',
        link_text='Firmwares',
        permissions=["netbox_firmware.view_firmware"],
        buttons= [
            PluginMenuButton(
                link='plugins:netbox_firmware:firmware_add',
                title='Add',
                icon_class='mdi mdi-plus-thick',
                permissions=["netbox_firmware.add_firmware"],
            ),
            PluginMenuButton(
                link='plugins:netbox_firmware:firmware_bulk_import',
                title='Import',
                icon_class='mdi mdi-upload',
                permissions=["netbox_firmware.add_firmware"],
            ),
        ],
    ),
    PluginMenuItem(
        link='plugins:netbox_firmware:firmwareassignment_list',
        link_text='Firmware Assignments',
        permissions=["netbox_firmware.view_firmwareassignment"],
        buttons= [
            PluginMenuButton(
                link='plugins:netbox_firmware:firmwareassignment_add',
                title='Add',
                icon_class='mdi mdi-plus-thick',
                permissions=["netbox_firmware.add_firmwareassignment"],
            ),
            PluginMenuButton(
                link='plugins:netbox_firmware:firmwareassignment_bulk_import',
                title='Import',
                icon_class='mdi mdi-upload',
                permissions=["netbox_firmware.add_firmwareassignment"],
            ),
        ],
    ),
)


### BIOS ###

bios_items = (
    PluginMenuItem(
        link='plugins:netbox_firmware:bios_list',
        link_text='BIOS',
        permissions=["netbox_firmware.view_bios"],
        buttons= [
            PluginMenuButton(
                link='plugins:netbox_firmware:bios_add',
                title='Add',
                icon_class='mdi mdi-plus-thick',
                permissions=["netbox_firmware.add_bios"],
            ),
            PluginMenuButton(
                link='plugins:netbox_firmware:bios_bulk_import',
                title='Import',
                icon_class='mdi mdi-upload',
                permissions=["netbox_firmware.add_bios"],
            ),
        ],
    ),
    PluginMenuItem(
        link='plugins:netbox_firmware:biosassignment_list',
        link_text='BIOS Assignments',
        permissions=["netbox_firmware.view_biosassignment"],
        buttons= [
            PluginMenuButton(
                link='plugins:netbox_firmware:biosassignment_add',
                title='Add',
                icon_class='mdi mdi-plus-thick',
                permissions=["netbox_firmware.add_biosassignment"],
            ),
            PluginMenuButton(
                link='plugins:netbox_firmware:biosassignment_bulk_import',
                title='Import',
                icon_class='mdi mdi-upload',
                permissions=["netbox_firmware.add_biosassignment"],
            ),
        ],
    ),
)

### -> Add new Menu Items above this line if needed.

### Menu ###


if get_plugin_config('netbox_firmware', 'top_level_menu'):
    menu = PluginMenu(
        label=f'Firmwares',
        groups=(
            ('Firmware', firmware_items),
            ('BIOS', bios_items),
        ),
        icon_class = 'mdi mdi-clipboard-text-multiple-outline'
    )
else:
    menu_items = firmware_items + bios_items


# Generated by Django 5.2.1 on 2025-09-10 11:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dcim', '0207_remove_redundant_indexes'),
        ('netbox_firmware', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='firmware',
            options={'ordering': ('name', 'module_type')},
        ),
        migrations.RemoveConstraint(
            model_name='firmware',
            name='firmware_device_type_or_module_type_required',
        ),
        migrations.AlterUniqueTogether(
            name='firmware',
            unique_together=set(),
        ),
        migrations.RemoveField(
            model_name='firmware',
            name='device_type',
        ),
        migrations.AddField(
            model_name='firmware',
            name='device_type',
            field=models.ManyToManyField(blank=True, null=True, related_name='firmware', to='dcim.devicetype'),
        ),
    ]



from django import forms
from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import gettext_lazy as _

from netbox_firmware.models import *
from netbox_firmware.choices import *
from netbox.choices import *
from netbox.forms import NetBoxModelImportForm
from utilities.forms.fields import (
    CSVChoiceField, CSVContentTypeField, CSVModelChoiceField, CSVModelMultipleChoiceField, CSVTypedChoiceField,
    SlugField,
)

### BIOS ###


class BiosImportForm(NetBoxModelImportForm):
    name = forms.CharField(
        label=_('Name'),
        required=True,
        help_text=_('Name of the BIOS')
    )
    manufacturer = CSVModelChoiceField(
        label=_('Manufacturer'),
        queryset=Manufacturer.objects.all(),
        to_field_name='name',
        required=True,
        help_text=_('BIOS type manufacturer')
    )
    file_name = forms.CharField(
        label=_('File name'),
        required=False,
        help_text=_('File name of the BIOS')
    )
    device_type = CSVModelMultipleChoiceField(
        label=_('Device Type name'),
        queryset=DeviceType.objects.all(),
        to_field_name='model',
        required=False,
        help_text=_('Name of the Device Type')
    )
    module_type = CSVModelMultipleChoiceField(
        label=_('Module Type name'),
        queryset=ModuleType.objects.all(),
        to_field_name='model',
        required=False,
        help_text=_('Name of the Module Type')
    )
    status = CSVChoiceField(
        label=_('Status'),
        choices=BiosStatusChoices,
        required=False,
        help_text=_('Operational status')
    )
    description = forms.CharField(
        label=_('Description'),
        required=False,
        help_text=_('Description of the BIOS')
    )
    comments = forms.CharField(
        label=_('Comments'),
        required=False,
        help_text=_('Additional comments about the BIOS')
    )

    class Meta:
        model = Bios
        fields = [
            'name', 
            'manufacturer',
            'file_name', 
            'status', 
            'description', 
            'device_type',
            'module_type',
            'comments',
            ]


    ### Check if only 1 type field is used.
    def clean(self):
        super().clean()
        device_type = self.cleaned_data.get('device_type')
        module_type = self.cleaned_data.get('module_type')

        # Ensure only one of device_type or module_type is set, not both
        if device_type and module_type:
            raise forms.ValidationError(
                "You cannot specify both Device Type and Module Type for a single BIOS entry."
            )
    
    ### This will validate that all listed DeviceType belong to the selected Manufacturer.
    def clean_device_type(self):
        device_type_list = self.cleaned_data.get('device_type')
        manufacturer = self.cleaned_data.get('manufacturer')

        # 1st condition : If list is empty = No validation needed
        if not device_type_list:
            return device_type_list

        # 2th condition : Check if all device_types belong to the selected manufacturer.
        if device_type_list and manufacturer:
            mismatched = [
                dt for dt in device_type_list if dt.manufacturer != manufacturer
            ]
            if mismatched:
                mismatched_names = ", ".join(str(dt) for dt in mismatched)
                raise forms.ValidationError(
                    f"The following DeviceTypes do not belong to the selected manufacturer '{manufacturer}': {mismatched_names}"
                )

        return device_type_list

   

    ### This will validate that all listed ModuleType belong to the selected Manufacturer.
    def clean_module_type(self):
        module_type_list = self.cleaned_data.get('module_type')
        manufacturer = self.cleaned_data.get('manufacturer')

        # 1st condition : If list is empty = No validation needed
        if not module_type_list:
            return module_type_list 

        # 2th condition : Check if all module_types belong to the selected manufacturer.
        if module_type_list and manufacturer:
            mismatched = [
                mt for mt in module_type_list if mt.manufacturer != manufacturer
            ]
            if mismatched:
                mismatched_names = ", ".join([str(mt) for mt in mismatched])
                raise forms.ValidationError(
                    f"The following ModuleTypes do not belong to the selected manufacturer '{manufacturer}': {mismatched_names}"
                )

        return module_type_list


### BIOSAssignment ###


class BiosAssignmentImportForm(NetBoxModelImportForm):
    bios = CSVModelChoiceField(
        label=_('Bios'),
        queryset=Bios.objects.all(),
        to_field_name='name',
        help_text=_('Bios name')
    )
    manufacturer= CSVModelChoiceField(
        queryset=Manufacturer.objects.all(),
        required=True,
        to_field_name='name',
        help_text=_('Manufacturer name')
    )
    hardware_kind = CSVTypedChoiceField(
        label=_('Hardware kind'),
        choices=HardwareKindChoices,
        required=True,
        help_text=_('Type of hardware')
    )
    hardware_name = forms.CharField(
        label=_('Hardware name'),
        required=True,
        help_text=_('Name of the hardware, e.g. device name or module id/serial')
    )
    comments = forms.CharField(
        label=_('Comments'),
        required=False,
        help_text=_('Additional comments about the bios assignment')
    )
    patch_date = forms.DateField(
        label=_('Patch date'),
        required=False,
        help_text=_('Date of the bios patch')
    )
    ticket_number = forms.CharField(
        label=_('Ticket number'),
        required=False,
        help_text=_('Ticket number of the bios patch')
    )
    description = forms.CharField(
        label=_('Description'),
        required=False,
        help_text=_('Description of the bios assignment')
    )

    class Meta:
        model = BiosAssignment
        fields = [
            'bios', 
            'manufacturer', 
            'hardware_kind', 
            'hardware_name', 
            'patch_date', 
            'ticket_number', 
            'description',
            'comments',
            ]

    def clean(self):
        super().clean()
        pass
    
    def _clean_fields(self):
        return super()._clean_fields()

    def _get_validation_exclusions(self):
        exclude = super()._get_validation_exclusions()
        exclude.remove('device')
        exclude.remove('module')
        return exclude

    def clean_hardware_name(self):
        hardware_kind = self.cleaned_data.get('hardware_kind')
        manufacturer = self.cleaned_data.get('manufacturer')
        model = self.cleaned_data.get('hardware_name')

        if not hardware_kind or not manufacturer:
            return None

        try:
            if hardware_kind == 'device':
                hardware_type = Device.objects.get(
                    device_type__manufacturer=manufacturer, name=model
                )
                existing = BiosAssignment.objects.filter(device__name=model).first()
                if existing and existing.id != self.instance.id:
                    raise ValidationError(f'Device "{model}" already has a BIOS assigned.')

            elif hardware_kind == 'module':
                if model.isdigit():
                    hardware_type = Module.objects.get(
                        module_type__manufacturer=manufacturer, pk=model
                    )
                    existing = BiosAssignment.objects.filter(module__pk=model).first()
                else:
                    hardware_type = Module.objects.get(
                        module_type__manufacturer=manufacturer, serial=model
                    )
                    existing = BiosAssignment.objects.filter(module__serial=model).first()

                if existing and existing.id != self.instance.id:
                    raise ValidationError(f'Module "{model}" already has a BIOS assigned.')

        except ObjectDoesNotExist:
            raise forms.ValidationError(
                f'Hardware type not found: "{hardware_kind}", "{manufacturer}", "{model}"'
            )

        setattr(self.instance, f'{hardware_kind}', hardware_type)
        return hardware_type

    def _get_clean_value(self, field_name):
        try:
            return self.base_fields[field_name].clean(self.data.get(field_name))
        except forms.ValidationError as e:
            self.add_error(field_name, e)
            raise


from django.template import Template
from django.db.models import Count, F
from netbox.plugins import PluginTemplateExtension

from .models import Firmware, FirmwareAssignment, Bios, BiosAssignment
from .utils import query_located

### Firmware ###

### This will show all related Firmware for the manufacturer in the manufacturer view, grouped by Devices and Modules.
class FirmwareCountsManufacturer(PluginTemplateExtension):
    models = ['dcim.manufacturer']
    def right_page(self):
        object = self.context.get('object')
        user = self.context['request'].user
        count_device = Firmware.objects.restrict(user, 'view').filter(device_type__manufacturer=object).distinct().count()
        count_module = Firmware.objects.restrict(user, 'view').filter(module_type__manufacturer=object).distinct().count()
        count_firmware = Firmware.objects.restrict(user, 'view').filter(manufacturer=object).distinct().count()
        context = {
            'firmware_stats': [
                {
                    'label': 'Devices',
                    'filter_field': 'manufacturer_id',
                    'extra_filter': '&kind=device',
                    'count': count_device,
                },
                {
                    'label': 'Modules',
                    'filter_field': 'manufacturer_id',
                    'extra_filter': '&kind=module',
                    'count': count_module,
                },
                {
                    'label': 'Total',
                    'filter_field': 'manufacturer_id',
                    'count': count_firmware,
                },
            ],
        }
        return self.render('netbox_firmware/inc/firmware_stats_counts.html', extra_context=context)

    

### This shows how many devices or modules that have this Firmware assigned in the Firmware view, grouped by Device Type and Module Type.
class FirmwareAssignmentsList(PluginTemplateExtension):
    models = ['netbox_firmware.firmware']

    def right_page(self):
        obj = self.context.get('object')

        # Devices
        qs_devices = (
            FirmwareAssignment.objects
            .filter(firmware=obj, device__isnull=False)
            .select_related('device__device_type')
        )
        device_type_counts = (
            qs_devices.values('device__device_type__id', 'device__device_type__model')
                .annotate(count=Count('id'))
                .order_by('device__device_type__model')
        )

        # Modules
        qs_modules = (
            FirmwareAssignment.objects
            .filter(firmware=obj, module__isnull=False)
            .select_related('module__module_type')
        )
        module_type_counts = (
            qs_modules.values('module__module_type__id', 'module__module_type__model')
                .annotate(count=Count('id'))
                .order_by('module__module_type__model')
        )

        context = {
            'count': qs_devices.count() + qs_modules.count(),
            'device_type_counts': device_type_counts,
            'module_type_counts': module_type_counts,
        }
        return self.render('netbox_firmware/inc/firmware_assignment_list.html', extra_context=context)



### This is the base class for showing assigned firmware in the device or module view.
### The actual classes are below.
class FirmwareAssignedInfoExtension(PluginTemplateExtension):
    def right_page(self):
        object = self.context.get('object')
        assignments = FirmwareAssignment.objects.filter(**{f'{self.kind}_id':object.id}).order_by('-patch_date')[:5]
        context = {
          'assignments': assignments
        }
        return self.render('netbox_firmware/inc/firmware_info.html', extra_context=context)

### This will show assigned firmware for the device in the device view.
class FirmwareAssignmentInfoDevice(FirmwareAssignedInfoExtension):
    models = ['dcim.device']
    kind = 'device'

### This will show assigned firmware for the module in the module view.
class FirmwareAssignmentInfoModule(FirmwareAssignedInfoExtension):
    models = ['dcim.module']
    kind = 'module'



### This will show all compatible firmwares for the device type in the device type view.
class FirmwareListDeviceType(PluginTemplateExtension):
    models = ['dcim.devicetype']

    def right_page(self):
        firmwares = self.context['object'].firmware.all()
        return self.render("netbox_firmware/inc/firmware_devicetype_list.html", extra_context={"firmwares": firmwares})



### This will show all compatible firmwares for the module type in the module type view.
class FirmwareListModuleType(PluginTemplateExtension):
    models = ['dcim.moduletype']

    def right_page(self):
        firmwares = self.context['object'].firmware.all()
        return self.render("netbox_firmware/inc/firmware_moduletype_list.html", extra_context={"firmwares": firmwares})



## ---------------------------------------- ##



### BIOS ###


### This will show all related BIOS for the manufacturer in the manufacturer view, grouped by Devices and Modules.
class BiosCountsManufacturer(PluginTemplateExtension):
    models = ['dcim.manufacturer']
    def right_page(self):
        object = self.context.get('object')
        user = self.context['request'].user
        count_device = Bios.objects.restrict(user, 'view').filter(device_type__manufacturer=object).distinct().count()
        count_module = Bios.objects.restrict(user, 'view').filter(module_type__manufacturer=object).distinct().count()
        count_bios = Bios.objects.restrict(user, 'view').filter(manufacturer=object).distinct().count()
        context = {
            'bios_stats': [
                {
                    'label': 'Devices',
                    'filter_field': 'manufacturer_id',
                    'extra_filter': '&kind=device',
                    'count': count_device,
                },
                {
                    'label': 'Modules',
                    'filter_field': 'manufacturer_id',
                    'extra_filter': '&kind=module',
                    'count': count_module,
                },
                {
                    'label': 'Total',
                    'filter_field': 'manufacturer_id',
                    'count': count_bios,
                },
            ],
        }
        return self.render('netbox_firmware/inc/bios_stats_counts.html', extra_context=context)



### This shows how many devices or modules that have this BIOS assigned in the BIOS view, grouped by Device Type and Module Type.
class BiosAssignmentsList(PluginTemplateExtension):
    models = ['netbox_firmware.bios']

    def right_page(self):
        obj = self.context.get('object')

        # Devices
        qs_devices = (
            BiosAssignment.objects
            .filter(bios=obj, device__isnull=False)
            .select_related('device__device_type')
        )
        device_type_counts = (
            qs_devices.values('device__device_type__id', 'device__device_type__model')
                .annotate(count=Count('id'))
                .order_by('device__device_type__model')
        )

        # Modules
        qs_modules = (
            BiosAssignment.objects
            .filter(bios=obj, module__isnull=False)
            .select_related('module__module_type')
        )
        module_type_counts = (
            qs_modules.values('module__module_type__id', 'module__module_type__model')
                .annotate(count=Count('id'))
                .order_by('module__module_type__model')
        )

        context = {
            'count': qs_devices.count() + qs_modules.count(),
            'device_type_counts': device_type_counts,
            'module_type_counts': module_type_counts,
        }
        return self.render('netbox_firmware/inc/bios_assignment_list.html', extra_context=context)



### This is the base class for showing assigned BIOS in the device or module view.
### The actual classes are below.
class BiosAssignedInfoExtension(PluginTemplateExtension):
    def right_page(self):
        object = self.context.get('object')
        assignments = BiosAssignment.objects.filter(**{f'{self.kind}_id':object.id}).order_by('-patch_date')[:5]
        context = {
          'assignments': assignments
        }
        return self.render('netbox_firmware/inc/bios_info.html', extra_context=context)

### This will show assigned BIOS for the device in the device view.
class BiosAssignmentInfoDevice(BiosAssignedInfoExtension):
    models = ['dcim.device']
    kind = 'device'

### This will show assigned BIOS for the module in the module view.
class BiosAssignmentInfoModule(BiosAssignedInfoExtension):
    models = ['dcim.module']
    kind = 'module'



### This will show all compatible BIOS for the device type in the device type view.
class BiosListDeviceType(PluginTemplateExtension):
    models = ['dcim.devicetype']

    def right_page(self):
        bios = self.context['object'].bios.all()
        return self.render("netbox_firmware/inc/bios_devicetype_list.html", extra_context={"bios": bios})



### This will show all compatible BIOS for the module type in the module type view.
class BiosListModuleType(PluginTemplateExtension):
    models = ['dcim.moduletype']

    def right_page(self):
        bios = self.context['object'].bios.all()
        return self.render("netbox_firmware/inc/bios_moduletype_list.html", extra_context={"bios": bios})


## ---------------------------------------- ##


template_extensions = (
    FirmwareCountsManufacturer,
    FirmwareListDeviceType,
    FirmwareListModuleType,
    FirmwareAssignmentsList,
    FirmwareAssignmentInfoDevice,
    FirmwareAssignmentInfoModule,
    BiosCountsManufacturer,
    BiosListDeviceType,
    BiosListModuleType,
    BiosAssignmentsList,
    BiosAssignmentInfoDevice,
    BiosAssignmentInfoModule,
)


from netbox.api.serializers import NetBoxModelSerializer
from dcim.api.serializers import ManufacturerSerializer, DeviceTypeSerializer, ModuleTypeSerializer, DeviceSerializer, ModuleSerializer
from netbox_firmware.models import Bios, BiosAssignment

__all__ = (
    'BiosSerializer',
    'BiosAssignmentSerializer',
)

class BiosSerializer(NetBoxModelSerializer):
    manufacturer = ManufacturerSerializer(nested=True, required=True)
    device_type = DeviceTypeSerializer(many=True, nested=True, required=False)
    module_type = ModuleTypeSerializer(many=True, nested=True, required=False)
    class Meta:
        model = Bios
        fields = '__all__'


class BiosAssignmentSerializer(NetBoxModelSerializer):
    bios = BiosSerializer(nested=True, required=True)
    device = DeviceSerializer(nested=True, required=False)
    module = ModuleSerializer(nested=True, required=False)
    class Meta:
        model = BiosAssignment
        fields = '__all__'


# Generated by Django 5.1.8 on 2025-06-02 08:55

import django.db.models.deletion
import django.db.models.functions.text
import taggit.managers
import utilities.json
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('dcim', '0200_populate_mac_addresses'),
        ('extras', '0123_journalentry_kind_default'),
    ]

    operations = [
        migrations.CreateModel(
            name='Bios',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('custom_field_data', models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder)),
                ('name', models.CharField(max_length=255)),
                ('file_name', models.CharField(blank=True, max_length=255, null=True)),
                ('file', models.FileField(blank=True, null=True, upload_to='bios-files')),
                ('status', models.CharField(default='active', max_length=50)),
                ('description', models.CharField(blank=True, max_length=255, null=True)),
                ('comments', models.TextField(blank=True, null=True)),
                ('device_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='bios', to='dcim.devicetype')),
                ('module_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='bios', to='dcim.moduletype')),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
            ],
            options={
                'verbose_name': 'BIOS',
                'verbose_name_plural': 'BIOS',
                'ordering': ('name', 'device_type', 'module_type'),
            },
        ),
        migrations.CreateModel(
            name='BiosAssignment',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('custom_field_data', models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder)),
                ('description', models.TextField(blank=True, null=True)),
                ('ticket_number', models.CharField(blank=True, max_length=100, null=True)),
                ('patch_date', models.DateField(blank=True, null=True)),
                ('comment', models.TextField(blank=True, null=True)),
                ('bios', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='BiosAssignment', to='netbox_firmware.bios')),
                ('device', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='BiosAssignment', to='dcim.device')),
                ('module', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='BiosAssignment', to='dcim.module')),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
            ],
            options={
                'verbose_name': 'BIOS Assignment',
                'verbose_name_plural': 'BIOS Assignments',
                'ordering': ('bios', 'device', 'module'),
            },
        ),
        migrations.CreateModel(
            name='Firmware',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('custom_field_data', models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder)),
                ('name', models.CharField(max_length=255)),
                ('file_name', models.CharField(blank=True, max_length=255, null=True)),
                ('file', models.FileField(blank=True, null=True, upload_to='firmware-files')),
                ('status', models.CharField(default='active', max_length=50)),
                ('description', models.CharField(blank=True, max_length=255, null=True)),
                ('comments', models.TextField(blank=True, null=True)),
                ('device_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='firmware', to='dcim.devicetype')),
                ('module_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='firmware', to='dcim.moduletype')),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
            ],
            options={
                'verbose_name': 'Firmware',
                'verbose_name_plural': 'Firmwares',
                'ordering': ('name', 'device_type', 'module_type'),
            },
        ),
        migrations.CreateModel(
            name='FirmwareAssignment',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('custom_field_data', models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder)),
                ('description', models.TextField(blank=True, null=True)),
                ('ticket_number', models.CharField(blank=True, max_length=100, null=True)),
                ('patch_date', models.DateField(blank=True, null=True)),
                ('comment', models.TextField(blank=True, null=True)),
                ('device', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='FirmwareAssignment', to='dcim.device')),
                ('firmware', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='FirmwareAssignment', to='netbox_firmware.firmware')),
                ('module', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='FirmwareAssignment', to='dcim.module')),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
            ],
            options={
                'verbose_name': 'Firmware Assignment',
                'verbose_name_plural': 'Firmware Assignments',
                'ordering': ('firmware', 'device', 'module'),
            },
        ),
        migrations.AddConstraint(
            model_name='bios',
            constraint=models.UniqueConstraint(django.db.models.functions.text.Lower('name'), name='netbox_firmware_bios_unique_name', violation_error_message='Device name must be unique.'),
        ),
        migrations.AddConstraint(
            model_name='bios',
            constraint=models.CheckConstraint(condition=models.Q(('device_type__isnull', False), ('module_type__isnull', False), _connector='OR'), name='bios_device_type_or_module_type_required'),
        ),
        migrations.AlterUniqueTogether(
            name='bios',
            unique_together={('name', 'device_type', 'module_type')},
        ),
        migrations.AddConstraint(
            model_name='biosassignment',
            constraint=models.CheckConstraint(condition=models.Q(('device__isnull', False), ('module__isnull', False), _connector='OR'), name='bios_device_or_module_required'),
        ),
        migrations.AddConstraint(
            model_name='biosassignment',
            constraint=models.UniqueConstraint(fields=('device',), name='unique_bios_per_device'),
        ),
        migrations.AddConstraint(
            model_name='biosassignment',
            constraint=models.UniqueConstraint(fields=('module',), name='unique_bios_per_module'),
        ),
        migrations.AddConstraint(
            model_name='firmware',
            constraint=models.UniqueConstraint(django.db.models.functions.text.Lower('name'), name='netbox_firmware_firmware_unique_name', violation_error_message='Device name must be unique.'),
        ),
        migrations.AddConstraint(
            model_name='firmware',
            constraint=models.CheckConstraint(condition=models.Q(('device_type__isnull', False), ('module_type__isnull', False), _connector='OR'), name='firmware_device_type_or_module_type_required'),
        ),
        migrations.AlterUniqueTogether(
            name='firmware',
            unique_together={('name', 'device_type', 'module_type')},
        ),
        migrations.AddConstraint(
            model_name='firmwareassignment',
            constraint=models.CheckConstraint(condition=models.Q(('device__isnull', False), ('module__isnull', False), _connector='OR'), name='firmassign_either_device_or_module_required'),
        ),
        migrations.AddConstraint(
            model_name='firmwareassignment',
            constraint=models.UniqueConstraint(fields=('device',), name='unique_firmware_per_device'),
        ),
        migrations.AddConstraint(
            model_name='firmwareassignment',
            constraint=models.UniqueConstraint(fields=('module',), name='unique_firmware_per_module'),
        ),
    ]



import strawberry_django

from netbox.graphql.filter_mixins import BaseFilterMixin

from netbox_firmware import filtersets, models

__all__ = (
    'FirmwareFilter',
    'FirmwareAssignmentFilter',
)

@strawberry_django.filter(models.Firmware, lookups=True)
class FirmwareFilter(BaseFilterMixin):
    pass

@strawberry_django.filter(models.FirmwareAssignment, lookups=True)
class FirmwareAssignmentFilter(BaseFilterMixin):
    pass

@strawberry_django.filter(models.Bios, lookups=True)
class BiosFilter(BaseFilterMixin):
    pass

@strawberry_django.filter(models.BiosAssignment, lookups=True)
class BiosAssignmentFilter(BaseFilterMixin):
    pass


import logging
from django.contrib import messages
from django.db import IntegrityError
from django.shortcuts import redirect
from django.template import Template
from utilities.query import count_related
from netbox.views import generic
from utilities.views import register_model_view

from netbox_firmware import filtersets
from netbox_firmware import tables
from netbox_firmware import forms
from netbox_firmware import models

__all__ = (
    'BiosAssignmentView',
    'BiosAssignmentListView',
)

@register_model_view(models.BiosAssignment)
class BiosAssignmentView(generic.ObjectView):
    queryset = models.BiosAssignment.objects.all()

    def get_extra_context(self, request, instance):
        context = super().get_extra_context(request, instance)
        return context

@register_model_view(models.BiosAssignment, 'list', path='', detail=False)
class BiosAssignmentListView(generic.ObjectListView):
    queryset = models.BiosAssignment.objects.prefetch_related(
        'device',
        'module',
        'bios',
    )
    filterset = filtersets.BiosAssignmentFilterSet
    filterset_form = forms.BiosAssignmentFilterForm
    table = tables.BiosAssignmentTable
    
@register_model_view(models.BiosAssignment, 'edit')
@register_model_view(models.BiosAssignment, 'add', detail=False)
class BiosAssignmentEditView(generic.ObjectEditView):
    queryset = models.BiosAssignment.objects.all()
    form = forms.BiosAssignmentForm
    default_return_url = 'plugins:netbox_firmware:biosassignment_list'

@register_model_view(models.BiosAssignment,'delete')
class BiosAssignmentDeleteView(generic.ObjectDeleteView):
    queryset = models.BiosAssignment.objects.all()
    default_return_url = 'plugins:netbox_firmware:biosassignment_list'
    
    def post(self, request, *args, **kwargs):
        return super().post(request, *args, **kwargs)

# ----------------- Bulk Import, Edit, Delete -----------------

@register_model_view(models.BiosAssignment, 'bulk_import', detail=False)
class BiosBulkImportView(generic.BulkImportView):
    queryset = models.BiosAssignment.objects.all()
    model_form = forms.BiosAssignmentImportForm

    def save_object(self, object_form, request):
        obj = object_form.save()
        return obj

@register_model_view(models.BiosAssignment, 'bulk_edit', detail=False)
class BiosAssignmentBulkEditView(generic.BulkEditView):
    queryset = models.BiosAssignment.objects.all()
    filterset = filtersets.BiosAssignmentFilterSet
    table = tables.BiosAssignmentTable
    form = forms.BiosAssignmentBulkEditForm
    default_return_url = 'plugins:netbox_firmware:biosassignment_list'
    
    def post (self, request, **kwargs):
        return super().post(request, **kwargs)

@register_model_view(models.BiosAssignment, 'bulk_delete', detail=False)
class BiosAssignmentBulkDeleteView(generic.BulkDeleteView):
    queryset = models.BiosAssignment.objects.all()
    table = tables.BiosAssignmentTable

    def post(self, request):
        return super().post(request)


import os
from netbox.plugins import PluginConfig
from django.conf import settings
from .version import __version__

class NetboxFirmwareConfig(PluginConfig):
    name = 'netbox_firmware'
    verbose_name = 'Netbox Firmware'
    version = __version__
    description = 'Firmware management in NetBox'
    author = 'Bart Van der Biest'
    author_email = 'bart@zimmo.be'
    base_url = 'firmware'
    min_version = '4.3.0'
    default_settings = {
        'top_level_menu': True,
    }

    def ready(self):
        firmware_dir = os.path.join(settings.MEDIA_ROOT, 'firmware-files')
        if not os.path.exists(firmware_dir):
            os.makedirs(firmware_dir)
        super().ready()
        from . import signals

config = NetboxFirmwareConfig


from django import forms
from django.utils.translation import gettext_lazy as _

from netbox_firmware.choices import BiosStatusChoices,HardwareKindChoices
from dcim.models import DeviceType, Manufacturer, ModuleType, Device, Module
from netbox.choices import *
from netbox.forms import NetBoxModelFilterSetForm
from utilities.forms import BOOLEAN_WITH_BLANK_CHOICES, FilterForm, add_blank_choice
from utilities.forms.fields import ColorField, DynamicModelMultipleChoiceField, TagFilterField
from utilities.forms.rendering import FieldSet
from utilities.forms.widgets import NumberWithOptions, DatePicker
from wireless.choices import *
from netbox_firmware.models import Bios, BiosAssignment

class BiosFilterForm(NetBoxModelFilterSetForm):
    model = Bios
    fieldsets = (
        FieldSet('q', 'tag', name=_('General')),
        FieldSet('manufacturer_id', 'status', name=_('BIOS')),
        FieldSet('kind', 'device_type_id', 'module_type_id', name=_('Hardware')),
    )

    selector_fields = ('q', 'status', 'manufacturer_id', 'device_type_id', 'module_type_id')

    manufacturer_id = DynamicModelMultipleChoiceField(
        queryset=Manufacturer.objects.all(),
        required=False,
        label=_('Manufacturer')
    )
    device_type_id = DynamicModelMultipleChoiceField(
        queryset=DeviceType.objects.all(),
        required=False,
        query_params={
            'manufacturer_id': '$manufacturer_id'
        },
        label=_('Device Type'),
    )
    module_type_id = DynamicModelMultipleChoiceField(
        queryset=ModuleType.objects.all(),
        required=False,
        query_params={
            'manufacturer_id': '$manufacturer_id'
        },
        label=_('Module Type'),
    )
    status = forms.MultipleChoiceField(
        label=_('Status'),
        choices=BiosStatusChoices,
        required=False
    )
    kind = forms.MultipleChoiceField(
        label=_('Kind'),
        choices=HardwareKindChoices,
        required=False
    )
    tag = TagFilterField(model)
    

class BiosAssignmentFilterForm(NetBoxModelFilterSetForm):
    model = BiosAssignment
    fieldsets = (
        FieldSet('q', 'tag'),
        FieldSet('patch_date',name=_('Patch Date')),
        FieldSet('manufacturer_id', 'kind', 'device_type_id', 'device_id', 'module_type_id', 'module_id', 'module_sn','module_device_id', name=_('Hardware')), 
        FieldSet('bios_id',name=_('BIOS')),
    )
    
    selector_fields = ('q', 'patch_date', 'manufacturer_id', 'device_id', 'module_id', 'bios_id')
    
    manufacturer_id = DynamicModelMultipleChoiceField(
        queryset=Manufacturer.objects.all(),
        required=False,
        label=_('Manufacturer')
    )
    kind = forms.MultipleChoiceField(
        label=_('Kind'),
        choices=HardwareKindChoices,
        required=False
    )
    device_type_id = DynamicModelMultipleChoiceField(
        queryset=DeviceType.objects.all(),
        required=False,
        label=_('Device Type'),
        query_params={
            'manufacturer_id': '$manufacturer_id'
        },
    )
    device_id = DynamicModelMultipleChoiceField(
        queryset=Device.objects.all(),
        required=False,
        query_params={
            'manufacturer_id': '$manufacturer_id',
            'device_type_id': '$device_type_id',
            'module_type_id': '$module_type_id'
        },
        label=_('Device')
    )
    module_type_id = DynamicModelMultipleChoiceField(
        queryset=ModuleType.objects.all(),
        required=False,
        label=_('Module Type'),
        query_params={
            'manufacturer_id': '$manufacturer_id'
        },
    )
    module_id = DynamicModelMultipleChoiceField(
        queryset=Module.objects.all(),
        required=False,
        query_params={
            'manufacturer_id': '$manufacturer_id',
            'device_type_id': '$device_type_id',
            'module_type_id': '$module_type_id'
        },
        label=_('Module')
    )
    module_sn = forms.CharField(
        required=False,
        label=_('Module Serial Number')
    )
    module_device_id = forms.ModelMultipleChoiceField(
        queryset=Device.objects.filter(modules__isnull=False).distinct(),
        required=False,
        label=_('Module Owner')
    )
    bios_id = DynamicModelMultipleChoiceField(
        queryset=Bios.objects.all(),
        required=False,
        query_params={
            'manufacturer_id': '$manufacturer_id',
            'device_id': '$device_id',
            'module_id': '$module_id'
        },
        label=_('BIOS'),
    )
    patch_date = forms.DateField(
        label=_('Patch Date'),
        required=False,
        widget=DatePicker()
    )
    tag = TagFilterField(model)


from netbox.api.viewsets import NetBoxModelViewSet
from .. import models
from .. import filtersets
from .serializers import (
    FirmwareSerializer, FirmwareAssignmentSerializer, BiosSerializer, BiosAssignmentSerializer
)

class FirmwareViewSet(NetBoxModelViewSet):
    queryset = models.Firmware.objects.all()
    serializer_class = FirmwareSerializer
    filterset_class = filtersets.FirmwareFilterSet

class FirmwareAssigmentViewSet(NetBoxModelViewSet):
    queryset = models.FirmwareAssignment.objects.all()
    serializer_class = FirmwareAssignmentSerializer
    filterset_class = filtersets.FirmwareAssignmentFilterSet
    
class BiosViewSet(NetBoxModelViewSet):
    queryset = models.Bios.objects.all()
    serializer_class = BiosSerializer
    filterset_class = filtersets.BiosFilterSet
    
class BiosAssigmentViewSet(NetBoxModelViewSet):
    queryset = models.BiosAssignment.objects.all()
    serializer_class = BiosAssignmentSerializer
    filterset_class = filtersets.BiosAssignmentFilterSet


import logging
import os
from django.db.models.signals import post_save, post_delete, pre_delete, pre_save
from django.dispatch import receiver

from netbox_firmware.models import *

### This is needed to delete the file from the storage when the Firmware or BIOS object is deleted or the file is changed. ###
@receiver(pre_save, sender=Bios)
@receiver(pre_save, sender=Firmware)
def delete_old_file_on_change(sender, instance, **kwargs):
    if not instance.pk:
        return

    try:
        old_instance = sender.objects.get(pk=instance.pk)
    except sender.DoesNotExist:
        return
    old_file = old_instance.file
    new_file = instance.file

    if not old_file:
        return

    if old_file and old_file != new_file:
        if os.path.isfile(old_file.path):
            os.remove(old_file.path)


# Generated by Django 5.2.1 on 2025-09-23 08:19

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dcim', '0207_remove_redundant_indexes'),
        ('netbox_firmware', '0006_alter_bios_options_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='bios',
            name='manufacturer',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='bios', to='dcim.manufacturer'),
        ),
    ]



from django.urls import include, path

from utilities.urls import get_model_urls
from netbox.views.generic import ObjectChangeLogView
from . import views

urlpatterns = [
    # Firmwares
    path('firmwares/', include(get_model_urls('netbox_firmware', 'firmware', detail=False))),
    path('firmwares/<int:pk>/', include(get_model_urls('netbox_firmware', 'firmware'))),
    
    # Assignments
    path('assignment/', include(get_model_urls('netbox_firmware','firmwareassignment',detail=False))),
    path('assignment/<int:pk>/', include(get_model_urls('netbox_firmware', 'firmwareassignment'))),
    
    # Bios
    path('bios/', include(get_model_urls('netbox_firmware', 'bios', detail=False))),
    path('bios/<int:pk>/', include(get_model_urls('netbox_firmware', 'bios'))),
    
    # Bios Assignments
    path('biosassignment/', include(get_model_urls('netbox_firmware','biosassignment',detail=False))),
    path('biosassignment/<int:pk>/', include(get_model_urls('netbox_firmware', 'biosassignment'))),
]


import strawberry
import strawberry_django

from netbox_firmware.models import (
    Firmware,
    FirmwareAssignment,
    Bios,
    BiosAssignment,
)
from .types import (
    FirmwareType,
    FirmwareAssignmentType,
    BiosType,
    BiosAssignmentType,
)

@strawberry.type
class FirmwareQuery:
    @strawberry.field
    def firmware(self, info, id: int) -> FirmwareType:
        return Firmware.objects.get(pk=id)
    firmware_list: list[FirmwareType] = strawberry_django.field()

@strawberry.type
class FirmwareAssignmentQuery:
    @strawberry.field
    def firmware_assignment(self, info, id: int) -> FirmwareAssignmentType:
        return FirmwareAssignment.objects.get(pk=id)
    firmware_assignment_list: list[FirmwareAssignmentType] = strawberry_django.field()

@strawberry.type
class BiosQuery:
    @strawberry.field
    def bios(self, info, id: int) -> BiosType:
        return Bios.objects.get(pk=id)
    bios_list: list[BiosType] = strawberry_django.field()

@strawberry.type
class BiosAssignmentQuery:
    @strawberry.field
    def bios_assignment(self, info, id: int) -> BiosAssignmentType:
        return BiosAssignment.objects.get(pk=id)
    bios_assignment_list: list[BiosAssignmentType] = strawberry_django.field()


import logging
from django.contrib import messages
from django.db import IntegrityError
from django.db.models import Count, Q
from django.shortcuts import redirect
from django.template import Template
from netbox.views import generic
from utilities.query import count_related
from utilities.views import register_model_view, ViewTab

from ..utils import get_countdevice
from .. import tables
from .. import forms
from .. import models
from .. import filtersets

__all__ = (
    'FirmwareView',
    'FirmwareListView',
)

@register_model_view(models.Firmware)
class FirmwareView(generic.ObjectView):
    queryset = models.Firmware.objects.all()

    def get_extra_context(self, request, instance):
        context = super().get_extra_context(request, instance)
        return context

@register_model_view(models.Firmware, 'list', path='', detail=False)
class FirmwareListView(generic.ObjectListView):
    queryset = models.Firmware.objects.prefetch_related(
        'device_type',
        'module_type'
    ).annotate(
        instance_count=count_related(models.FirmwareAssignment,'firmware'),
    )
    filterset = filtersets.FirmwareFilterSet
    filterset_form = forms.FirmwareFilterForm
    table = tables.FirmwareTable

@register_model_view(models.Firmware, 'edit')
@register_model_view(models.Firmware, 'add', detail=False)
class FirmwareEditView(generic.ObjectEditView):
    queryset = models.Firmware.objects.all()
    form = forms.FirmwareForm
    default_return_url = 'plugins:netbox_firmware:firmware_list'

@register_model_view(models.Firmware,'delete')
class FirmwareDeleteView(generic.ObjectDeleteView):
    queryset = models.Firmware.objects.all()
    default_return_url = 'plugins:netbox_firmware:firmware_list'
    
    def post(self, request, *args, **kwargs):
        return super().post(request, *args, **kwargs)


# ----------------- Bulk Import, Edit, Delete -----------------

@register_model_view(models.Firmware, 'bulk_import', detail=False)
class FirmwareBulkImportView(generic.BulkImportView):
    queryset = models.Firmware.objects.all()
    model_form = forms.FirmwareImportForm

    def save_object(self, object_form, request):
        obj = object_form.save()
        return obj
    
@register_model_view(models.Firmware, 'bulk_edit', detail=False)
class FirmwareBulkEditView(generic.BulkEditView):
    queryset = models.Firmware.objects.all()
    filterset = filtersets.FirmwareFilterSet
    table = tables.FirmwareTable
    form = forms.FirmwareBulkEditForm
    default_return_url = 'plugins:netbox_firmware:firmware_list'
    
    def post_save_operations(self, form, obj):
        super().post_save_operations(form, obj)

        # Add/remove Linked Device Types
        if form.cleaned_data.get('add_device_type', None):
            obj.device_type.add(*form.cleaned_data['add_device_type'])
        if form.cleaned_data.get('remove_device_type', None):
            obj.device_type.remove(*form.cleaned_data['remove_device_type'])

        # Add/remove Linked Module Types
        if form.cleaned_data.get('add_module_type', None):
            obj.module_type.add(*form.cleaned_data['add_module_type'])
        if form.cleaned_data.get('remove_module_type', None):
            obj.module_type.remove(*form.cleaned_data['remove_module_type'])

    def post (self, request, **kwargs):
        return super().post(request, **kwargs)


@register_model_view(models.Firmware, 'bulk_delete', detail=False)
class FirmwareBulkDeleteView(generic.BulkDeleteView):
    queryset = models.Firmware.objects.all()
    table = tables.FirmwareTable

    def post(self, request):
        return super().post(request)


from typing import Annotated, List

import strawberry
import strawberry_django

from netbox.graphql.scalars import BigInt
from netbox.graphql.types import (
    NetBoxObjectType,
    OrganizationalObjectType,
)
from netbox_firmware.models import (
    Firmware,
    FirmwareAssignment,
    Bios,
    BiosAssignment
)
from .filters import (
    FirmwareFilter,
    FirmwareAssignmentFilter,
    BiosFilter,
    BiosAssignmentFilter
)

@strawberry_django.type(Firmware, fields='__all__', filters=FirmwareFilter)
class FirmwareType(NetBoxObjectType):
    manufacturer: Annotated["ManufacturerType", strawberry.lazy('dcim.graphql.types')]
    device_type: List[Annotated["DeviceTypeType", strawberry.lazy("dcim.graphql.types")]] | None
    module_type: List[Annotated["ModuleTypeType", strawberry.lazy('dcim.graphql.types')]] | None

@strawberry_django.type(FirmwareAssignment, fields='__all__', filters=FirmwareAssignmentFilter)
class FirmwareAssignmentType(NetBoxObjectType):
    firmware: Annotated["FirmwareType", strawberry.lazy("netbox_firmware.graphql.types")]
    module: Annotated["ModuleType", strawberry.lazy('dcim.graphql.types')] | None
    device: Annotated["DeviceType", strawberry.lazy('dcim.graphql.types')] | None

@strawberry_django.type(Bios, fields='__all__', filters=BiosFilter)
class BiosType(NetBoxObjectType):
    manufacturer: Annotated["ManufacturerType", strawberry.lazy('dcim.graphql.types')]
    device_type: List[Annotated["DeviceTypeType", strawberry.lazy("dcim.graphql.types")]] | None
    module_type: List[Annotated["ModuleTypeType", strawberry.lazy('dcim.graphql.types')]] | None

@strawberry_django.type(BiosAssignment, fields='__all__', filters=BiosAssignmentFilter)
class BiosAssignmentType(NetBoxObjectType):
    bios: Annotated["BiosType", strawberry.lazy("netbox_firmware.graphql.types")]
    module: Annotated["ModuleType", strawberry.lazy('dcim.graphql.types')] | None
    device: Annotated["DeviceType", strawberry.lazy('dcim.graphql.types')] | None


from django.urls import path, include
from netbox.api.routers import NetBoxRouter
from . import views

app_name = 'netbox_firmware'

router = NetBoxRouter()
router.register(r'firmware', views.FirmwareViewSet)
router.register(r'firmware-assignment', views.FirmwareAssigmentViewSet)

router.register(r'bios', views.BiosViewSet)
router.register(r'bios-assignment', views.BiosAssigmentViewSet)

urlpatterns = [
    path('', include(router.urls)),
]


from django.conf import settings
from django.db.models import Q
from django.db.models import Count, OuterRef, Subquery
from django.db.models.functions import Coalesce
from django.core.exceptions import ImproperlyConfigured
from django.db.models.signals import pre_save

from django.urls import reverse
from django.utils.html import format_html, format_html_join
from django_tables2 import tables
from dcim.models import Device, Module, Rack
from netbox.plugins import get_plugin_config
from .choices import FirmwareStatusChoices

class FirmwareColumn(tables.Column):
    def render(self, value, record):
        # `record` is een Device instance
        assignments = record.FirmwareAssignment.all()
        if not assignments:
            return "-"
        links = [
            format_html(
                '<a href="{}">{}</a>',
                reverse('plugins:netbox_firmware:firmwareassignment', args=[a.pk]),
                a.firmware
            )
            for a in assignments
        ]
        return format_html_join(", ", "{}", ((link,) for link in links))

class BiosColumn(tables.Column):
    def render(self, value, record):
        # `record` is een Device instance
        assignments = record.BiosAssignment.all()
        if not assignments:
            return "-"
        links = [
            format_html(
                '<a href="{}">{}</a>',
                reverse('plugins:netbox_firmware:biosassignment', args=[a.pk]),
                a.bios
            )
            for a in assignments
        ]
        return format_html_join(", ", "{}", ((link,) for link in links))

def get_prechange_field(obj, field_name):
    """Get value from obj._prechange_snapshot. If field is a relation,
    return object instance.
    """
    value = getattr(obj, '_prechange_snapshot', {}).get(field_name)
    if value is None:
        return None
    field = obj._meta.get_field(field_name)
    if field.is_relation:
        rel_obj = field.related_model.objects.filter(pk=value).first()
        if rel_obj:
            return rel_obj
        else:
            return None
    return value


def get_plugin_setting(setting_name):
    return get_plugin_config('netbox_firmware', setting_name)


def get_status_for(status):
    status_name = get_plugin_setting(status + '_status_name')
    if status_name is None:
        return None
    if status_name not in dict(FirmwareStatusChoices):
        raise ImproperlyConfigured(
            f'netbox_firmware plugin configuration defines status {status_name}, but it is not defined in FIELD_CHOICES["netbox_firmware.Firmware.status"]'
        )
    return status_name


def get_all_statuses_for(status):
    status_names = get_plugin_setting(status + '_additional_status_names')
    status_names = set(status_names)
    # add primary status
    if primary_status := get_status_for(status):
        status_names.add(primary_status)
    if len(status_names) < 1:
        return None
    if extra_statuses := status_names.difference(set(dict(FirmwareStatusChoices))):
        raise ImproperlyConfigured(
            f'netbox_firmware plugin configuration defines statuses {extra_statuses}, but these are not defined in FIELD_CHOICES["netbox_firmware.Firmware.status"]'
        )
    return list(status_names)


def get_tags_that_protect_firmware_from_deletion():
    """Return a list of tags that prevent an firmware from being deleted.

    Which tags prevent deletion is defined in the plugin configuration,
    under the key ``firmware_disable_deletion_for_tags``.

    Returns:
        list: list of tag slug strings
    """
    return get_plugin_setting('firmware_disable_deletion_for_tags')


def get_tags_and_edit_protected_firmware_fields():
    """Return a dict of tags and fields that prevent editing specified fields.

    Which tags prevent editing is defined in the plugin configuration,
    under the key ``firmware_disable_editing_fields_for_tags``.

    Structure of the dict is::

        {
            'tag_slug': ['field1', 'field2'],
            'tag_slug2': ['field1', 'field4'],
        }

    Returns:
        dict: dict of tag slug strings and list of field names
    """
    return get_plugin_setting('firmware_disable_editing_fields_for_tags')


def firmware_set_new_hw(firmware, hw):
    """
    Firmware was assigned to hardware (device/module/inventory item/rack) and we want to
    sync some field values from firmware to hardware
    Validation if firmware can be assigned to hw should be done before calling this function.
    """
    # device, module... needs None for blank firmware_tag to enforce uniqness at DB level
    new_firmware_tag = firmware.firmware_tag or None
    # device, module... does not allow serial to be null
    new_serial = firmware.serial or ''
    hw_save = False
    if hw.serial != new_serial:
        hw.serial = new_serial
        hw_save = True
    if hw.firmware_tag != new_firmware_tag:
        hw.firmware_tag = new_firmware_tag
        hw_save = True
    # handle changing of model (<kind>_type)
    if firmware.kind in ['device', 'module', 'rack']:
        firmware_type = getattr(firmware, firmware.kind+'_type')
        hw_type = getattr(hw, firmware.kind+'_type')
        if firmware_type != hw_type:
            setattr(hw, firmware.kind+'_type', firmware_type)
            hw_save = True
    if hw_save:
        hw.save()


def is_equal_none(a, b):
    """ Compare a and b as string. None is considered the same as empty string. """
    if a is None or b is None:
        return a == b or a == '' or b == ''
    return a == b


def query_located(queryset, field_name, values, firmwares_shown='all'):
    """
    Filters queryset on located values. Can filter for installed
    location/site and/or stored location/site for firmwares makred as stored.
    Args:
        * queryset - queryset of Firmware model
        * field_name - 'site' or 'location' or 'rack'
        * values - list of PKs of location types to filter on
        * firmwares_shown - 'all' or 'installed' or 'stored'
    """
    if field_name == 'rack':
        q_installed = Q(**{f'rack__in':values})
    else:
        q_installed = Q(**{f'rack__{field_name}__in':values})
    q_installed = (
        q_installed|
        Q(**{f'device__{field_name}__in':values})|
        Q(**{f'module__device__{field_name}__in':values})
    )

    # Q expressions for stored
    if field_name == 'rack':
        # storage in rack is not supported
        # generate Q() that matches none
        q_stored = Q(pk__in=[])
    elif field_name == 'location':
        q_stored = (
            Q(**{f'storage_location__in':values})&
            Q(status__in=get_all_statuses_for('stored'))
        )
    elif field_name == 'site':
        q_stored = (
            Q(**{f'storage_location__site__in':values})&
            Q(status__in=get_all_statuses_for('stored'))
        )

    if firmwares_shown == 'all':
        q = q_installed | q_stored
    elif firmwares_shown == 'installed':
        q = q_installed
    elif firmwares_shown == 'stored':
        q = q_stored
    else:
        raise Exception('unsupported')
    return queryset.filter(q)


def get_firmware_custom_fields_search_filters():
    """Returns a list of custom field filter strings that can be used in Q() filter.

    Custom fields and filters are used is defined in the plugin configuration,
    under the key ``firmware_custom_fields_search_filters``.

    Returns:
        list: list of custom field filter strings
    """
    custom_fields_filters = get_plugin_setting('firmware_custom_fields_search_filters')

    fields = []
    for field_name, filters in custom_fields_filters.items():
        for filter in filters:
            fields.append(f"custom_field_data__{field_name}__{filter}")
    return fields


def get_countdevice(model,modelfield,field):
    """
    Return a Subquery suitable for annotating a child object count.
    """
    subquery = Subquery(
        model.objects.filter(
            **{modelfield: OuterRef('pk')}
        ).order_by().values(
            modelfield
        ).annotate(
            c=Count(field)
        ).values('c')
    )

    return Coalesce(subquery, 0)



from netbox.search import SearchIndex, register_search
from . import models

@register_search
class FirmwareIndex(SearchIndex):
    model = models.Firmware
    fields = (
        ('name', 100),
        ('description', 500),
        ('comments', 5000),
    )
    display_attrs =(
        'status', 'manufacturer'
    )
    
@register_search
class FirmwareAssignmentIndex(SearchIndex):
    model = models.FirmwareAssignment
    fields = (
        ('firmware', 100),
        ('device', 100),
        ('module', 100),
        ('description',500),
        ('comment', 5000),
    )
    display_attrs = (
        'firmware', 'device', 'module'
    )

@register_search
class BiosIndex(SearchIndex):
    model = models.Bios
    fields = (
        ('name', 100),
        ('description', 500),
        ('comments', 5000),
    )
    display_attrs = (
        'status', 'manufacturer'
    )

@register_search
class BiosAssignmentIndex(SearchIndex):
    model = models.BiosAssignment
    fields = (
        ('bios', 100),
        ('device', 100),
        ('module', 100),
        ('description', 500),
        ('comment', 5000),
    )
    display_attrs = (
        'bios', 'device', 'module'
    )


from django import forms
from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import gettext_lazy as _

from netbox_firmware.models import *
from netbox_firmware.choices import *
from netbox.choices import *
from netbox.forms import NetBoxModelImportForm
from utilities.forms.fields import (
    CSVChoiceField, CSVContentTypeField, CSVModelChoiceField, CSVModelMultipleChoiceField, CSVTypedChoiceField,
    SlugField,
)

### Firmware ###


class FirmwareImportForm(NetBoxModelImportForm):
    name = forms.CharField(
        label=_('Name'),
        required=True,
        help_text=_('Name of the firmware')
    )
    manufacturer = CSVModelChoiceField(
        label=_('Manufacturer'),
        queryset=Manufacturer.objects.all(),
        to_field_name='name',
        required=True,
        help_text=_('Firmware type manufacturer')
    )
    file_name = forms.CharField(
        label=_('File name'),
        required=False,
        help_text=_('File name of the firmware')
    )
    device_type = CSVModelMultipleChoiceField(
        label=_('Device Type name'),
        queryset=DeviceType.objects.all(),
        to_field_name='model',
        required=False,
        help_text=_('Name of the Device Type. / List if the Firmware supports multiple Device Type')
    )
    module_type = CSVModelMultipleChoiceField(
        label=_('Module Type name'),
        queryset=ModuleType.objects.all(),
        to_field_name='model',
        required=False,
        help_text=_('Name of the Module Type. / List if the Firmware supports multiple Module Type')
    )
    status = CSVChoiceField(
        label=_('Status'),
        choices=FirmwareStatusChoices,
        required=False,
        help_text=_('Operational status. / Default = Active')
    )
    description = forms.CharField(
        label=_('Description'),
        required=False,
        help_text=_('Description of the firmware')
    )
    comments = forms.CharField(
        label=_('Comments'),
        required=False,
        help_text=_('Additional comments about the firmware')
    )

    class Meta:
        model = Firmware
        fields = [
            'name', 
            'manufacturer',
            'file_name', 
            'status', 
            'description', 
            'device_type',
            'module_type',
            'comments',
            ]


    ### Check if only 1 type field is used.
    def clean(self):
        super().clean()
        device_type = self.cleaned_data.get('device_type')
        module_type = self.cleaned_data.get('module_type')

        # Ensure only one of device_type or module_type is set, not both
        if device_type and module_type:
            raise forms.ValidationError(
                "You cannot specify both Device Type and Module Type for a single firmware entry."
            )

    ### This will validate that all listed DeviceType belong to the selected Manufacturer.
    def clean_device_type(self):
        device_type_list = self.cleaned_data.get('device_type')
        manufacturer = self.cleaned_data.get('manufacturer')

        # 1st condition : If list is empty = No validation needed.
        if not device_type_list:
            return device_type_list

        # 2th condition : Check if all device_types belong to the selected manufacturer.
        if device_type_list and manufacturer:
            mismatched = [
                dt for dt in device_type_list if dt.manufacturer != manufacturer
            ]
            if mismatched:
                mismatched_names = ", ".join(str(dt) for dt in mismatched)
                raise forms.ValidationError(
                    f"The following DeviceTypes do not belong to the selected manufacturer '{manufacturer}': {mismatched_names}"
                )

        return device_type_list

   

    ### This will validate that all listed ModuleType belong to the selected Manufacturer.
    def clean_module_type(self):
        module_type_list = self.cleaned_data.get('module_type')
        manufacturer = self.cleaned_data.get('manufacturer')

        # 1st condition : If list is empty = No validation needed.
        if not module_type_list:
            return module_type_list 

        # 2th condition : Check if all module_types belong to the selected manufacturer.
        if module_type_list and manufacturer:
            mismatched = [
                mt for mt in module_type_list if mt.manufacturer != manufacturer
            ]
            if mismatched:
                mismatched_names = ", ".join([str(mt) for mt in mismatched])
                raise forms.ValidationError(
                    f"The following ModuleTypes do not belong to the selected manufacturer '{manufacturer}': {mismatched_names}"
                )

        return module_type_list


### FirmwareAssignment ###


class FirmwareAssignmentImportForm(NetBoxModelImportForm):
    firmware = CSVModelChoiceField(
        label=_('Firmware'),
        queryset=Firmware.objects.all(),
        to_field_name='name',
        help_text=_('Firmware name')
    )
    manufacturer = CSVModelChoiceField(
        label=_('Manufacturer'),
        queryset=Manufacturer.objects.all(),
        to_field_name='name',
        help_text=_('Device type manufacturer')
    )
    hardware_kind = CSVTypedChoiceField(
        label=_('Hardware kind'),
        choices=HardwareKindChoices,
        required=True,
        help_text=_('Type of hardware')
    )
    hardware_name = forms.CharField(
        label=_('Hardware name'),
        required=True,
        help_text=_('Name of the hardware, e.g. device name or module id/serial')
    )
    comments = forms.CharField(
        label=_('Comments'),
        required=False,
        help_text=_('Additional comments about the firmware assignment')
    )
    patch_date = forms.DateField(
        label=_('Patch date'),
        required=False,
        help_text=_('Date of the firmware patch')
    )
    ticket_number = forms.CharField(
        label=_('Ticket number'),
        required=False,
        help_text=_('Ticket number of the firmware patch')
    )
    description = forms.CharField(
        label=_('Description'),
        required=False,
        help_text=_('Description of the firmware assignment')
    )

    class Meta:
        model = FirmwareAssignment
        fields = [
            'firmware', 
            'manufacturer', 
            'hardware_kind', 
            'hardware_name', 
            'patch_date', 
            'ticket_number', 
            'description',
            'comments',
            ]

    def clean(self):
        super().clean()
        pass
    
    def _clean_fields(self):
        return super()._clean_fields()

    def _get_validation_exclusions(self):
        exclude = super()._get_validation_exclusions()
        exclude.remove('device')
        exclude.remove('module')
        return exclude

    def clean_hardware_name(self):
        hardware_kind = self.cleaned_data.get('hardware_kind')
        manufacturer = self.cleaned_data.get('manufacturer')
        model = self.cleaned_data.get('hardware_name')

        if not hardware_kind or not manufacturer:
            return None

        try:
            if hardware_kind == 'device':
                hardware_type = Device.objects.get(
                    device_type__manufacturer=manufacturer, name=model
                )
                existing = FirmwareAssignment.objects.filter(device__name=model).first()
                if existing and existing.id != self.instance.id:
                    raise ValidationError(f'Device "{model}" already has a Firmware assigned.')

            elif hardware_kind == 'module':
                if model.isdigit():
                    hardware_type = Module.objects.get(
                        module_type__manufacturer=manufacturer, pk=model
                    )
                    existing = FirmwareAssignment.objects.filter(module__pk=model).first()
                else:
                    hardware_type = Module.objects.get(
                        module_type__manufacturer=manufacturer, serial=model
                    )
                    existing = FirmwareAssignment.objects.filter(module__serial=model).first()

                if existing and existing.id != self.instance.id:
                    raise ValidationError(f'Module "{model}" already has a Firmware assigned.')

        except ObjectDoesNotExist:
            raise forms.ValidationError(
                f'Hardware type not found: "{hardware_kind}", "{manufacturer}", "{model}"'
            )

        setattr(self.instance, f'{hardware_kind}', hardware_type)
        return hardware_type

    def _get_clean_value(self, field_name):
        try:
            return self.base_fields[field_name].clean(self.data.get(field_name))
        except forms.ValidationError as e:
            self.add_error(field_name, e)
            raise


from django_tables2 import tables
from django.utils.translation import gettext_lazy as _

from netbox.tables import NetBoxTable, columns
from ..models import Bios, BiosAssignment
from netbox_firmware.utils import BiosColumn

from dcim.tables import DeviceTable, ModuleTable
from utilities.tables import register_table_column

from .template_code import *

__all__ = (
    'BiosTable',
    'BiosAssignmentTable',
)

class BiosTable(NetBoxTable):
    name = tables.Column(
        linkify=True,
    )
    description = tables.Column()
    file_name = tables.Column()
    comments = tables.Column()
    status = tables.Column()
    manufacturer = tables.Column(
        verbose_name=_('Manufacturer'),
        linkify=True,
    )
    device_type = columns.TemplateColumn(
        template_code=SHOW_LINKED_DEVICE_TYPES,
        linkify=True,
        )
    module_type = columns.TemplateColumn(
        template_code=SHOW_LINKED_MODULE_TYPES,
        linkify=True,
        )
    instance_count = columns.LinkedCountColumn(
        viewname='plugins:netbox_firmware:biosassignment_list',
        url_params={'bios_id': 'pk'},
        verbose_name=_('Instances')
    )
    file_path = tables.Column(
        accessor='file_path',
        verbose_name=_('File path'),
        orderable=False
    )

    class Meta(NetBoxTable.Meta):
        model = Bios
        fields = ('name', 'file_name', 'comments', 'status', 
                  'manufacturer', 
                  'module_type', 'device_type'
                  )
        #░ Default columns moeten nog gedefinieerd worden in de Meta class

    # def order_manufacturer(self, queryset, is_descending):
    #     if is_descending:
    #         ordering_device = '-device_type__manufacturer'
    #         ordering_module = '-module_type__manufacturer'
    #     else:
    #         ordering_device = 'device_type__manufacturer'
    #         ordering_module = 'module_type__manufacturer'
    
    #     # Voeg de twee velden toe aan de query voor een gecombineerde sortering
    #     queryset = queryset.order_by(ordering_device, ordering_module)
    #     return queryset, True

class BiosAssignmentTable(NetBoxTable):
    description = tables.Column()
    ticket_number = tables.Column()
    patch_date = tables.Column()
    bios = tables.Column(accessor='bios',verbose_name='BIOS',linkify=True,)
    module = tables.Column(accessor='module',verbose_name="Module",linkify=True,)
    module_device= tables.Column(accessor='module_device',verbose_name='Module owner',linkify=True)
    device = tables.Column(accessor='device',verbose_name="Device",linkify=True,)
    manufacturer = tables.Column(
        verbose_name=_('Manufacturer'),
        accessor='get_manufacturer',  # We gebruiken de method get_manufacturer
        linkify=True,
    )
    
    device_type = tables.Column(accessor='device_type',verbose_name='Device Type',linkify=True)
    device_sn = tables.Column(accessor='device_sn',verbose_name='Device Serial Number')
    module_type = tables.Column(accessor='module_type',verbose_name='Module Type',linkify=True)
    module_sn = tables.Column(accessor='module_sn',verbose_name='Module Serial Number')
    
    class Meta(NetBoxTable.Meta):
        model = BiosAssignment
        fields = ('description','ticket_number','patch_date',
                  'bios','device', 'module'
                  )

    # Order methods
    def order_manufacturer(self, queryset, is_descending):
        if is_descending:
            ordering_device = '-device__device_type__manufacturer'
            ordering_module = '-module__module_type__manufacturer'
        else:
            ordering_device = 'device__device_type__manufacturer'
            ordering_module = 'module__module_type__manufacturer'
    
        # Voeg de twee velden toe aan de query voor een gecombineerde sortering
        queryset = queryset.order_by(ordering_device, ordering_module)
        return queryset, True

    def order_device_sn(self, queryset, is_descending):
        ordering = ('-device__serial' if is_descending else 'device__serial')
        queryset = queryset.order_by(ordering)
        return queryset, True

    def order_module_device(self, queryset, is_descending):
        ordering = ('-module__device' if is_descending else 'module__device')
        queryset = queryset.order_by(ordering)
        return queryset, True
    
    def order_module_sn(self, queryset, is_descending):
        ordering = ('-module__serial' if is_descending else 'module__serial')
        queryset = queryset.order_by(ordering)
        return queryset, True

    def order_device_type(self, queryset, is_descending):
        ordering = ('-device__device_type__model' if is_descending else 'device__device_type__model')
        queryset = queryset.order_by(ordering)
        return queryset, True

    def order_module_type(self, queryset, is_descending):
        ordering = ('-module__module_type__model' if is_descending else 'module__module_type__model')
        queryset = queryset.order_by(ordering)
        return queryset, True

# ========================
# DCIM model table columns
# ========================


### Add BIOS column to Device and Module tables.
bios_column = BiosColumn(
    verbose_name=_('BIOS'),
    orderable=True,
)

register_table_column(bios_column, 'BiosAssignment', DeviceTable)
register_table_column(bios_column, 'BiosAssignment', ModuleTable)